#!/usr/bin/env python3

import sys
from argparse import ArgumentParser

from ..core import dataset
from ..core.load import load_all
from ._common import commonparser, iter_dir


def to_netcdf(read: str, name: str, quiet: bool = True) -> None:
    if not quiet:
        print(f"Loading {read} ...", end="")
        sys.stdout.flush()
    ds = load_all(read)
    if not quiet:
        print(" writing ...", end="")
        sys.stdout.flush()
    ds.emc3.to_netcdf(name + ".nc")
    if not quiet:
        print(" done")


def parser() -> ArgumentParser:
    return commonparser(
        "Load the data from EMC3 simulations and store as netcdf file. "
        "The data is written for each simulation to a netcdf file."
    )


def main() -> None:
    args = parser().parse_args()

    for d, n in iter_dir(args):
        to_netcdf(d, n, args.quiet)


if __name__ == "__main__":
    main()



import tempfile

import xarray as xr
from hypothesis import assume, given, settings
from hypothesis import strategies as st

from .. import load, write
from ..cli.append_time import append_time as append
from . import gen_ds as g
from .test_write_load import assert_ds_are_equal


@settings(deadline=None)
@given(
    g.hypo_shape(100),
    g.hypo_vars12(),
    st.integers(min_value=1, max_value=10),
)
def disabled_test_average(shape, v12, rep):
    v1, v2 = v12
    assume(len(v2))
    org = g.gen_rand(shape, v1)
    orgs = [org]
    read = []
    with tempfile.TemporaryDirectory() as dir:
        write.fortran(org, dir)
        append(dir)
        for i in range(rep):
            d2 = g.gen_updated(org, v2)
            orgs.append(d2)
            write.fortran(d2, dir)
            append(dir)
        read = xr.open_dataset(f"{dir}.nc")
        for i, o in enumerate(orgs):
            assert_ds_are_equal(o, read.isel(time=i), True, 1e-2, 1e-2)

        with tempfile.TemporaryDirectory() as dir2:
            write.fortran(read.emc3.average_time(), dir2)
            dn = load.all(dir2)
        assert_ds_are_equal(read.emc3.average_time(), dn, True, 1e-4)



#!/usr/bin/env python3

import os
import sys
from argparse import ArgumentParser

import xarray as xr

from ..core.load import archive, load_all
from ._common import commonparser, iter_dir

ext = "arch.nc"


def to_archive(
    read: str, name: str, quiet: bool, geom: bool, mapping: bool, delete: bool
) -> None:
    fromds = False
    if not quiet:
        print(f"Loading {read} ...", end="")
        sys.stdout.flush()
    if os.path.isdir(read):
        ds = load_all(read)
    else:
        ds = xr.open_dataset(read)
        fromds = True
    if not quiet:
        print(" writing ...", end="")
        sys.stdout.flush()
    archive(ds, f"{name}.{ext}", geom, mapping)
    if delete and fromds:
        if not quiet:
            print(" deleting ...", end="")
            sys.stdout.flush()
        os.unlink(read)
    if not quiet:
        print(" done")


def parser() -> ArgumentParser:
    parser = commonparser(
        "Archive the data from EMC3 simulations as a netcdf file.", ext=ext
    )
    parser.add_argument(
        "-g", "--geometry", action="store_true", help="Also store geometry"
    )
    parser.add_argument(
        "-n",
        "--no-mapping",
        action="store_true",
        help="Do not include the mapping information.",
    )
    parser.add_argument(
        "-d",
        "--delete",
        action="store_true",
        help="Delete the uncompressed input file. "
        "Only used if the input was a netcdf file",
    )
    return parser


def main() -> None:
    args = parser().parse_args()

    for d, n in iter_dir(args):
        to_archive(
            d,
            n,
            args.quiet,
            geom=args.geometry,
            mapping=not args.no_mapping,
            delete=args.delete,
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import typing

import numpy as np
import xarray as xr
from hypothesis import assume
from hypothesis import strategies as st

from ..core import dataset, load, utils

dims = "r", "theta", "phi"

setting = dict(deadline=None, max_examples=10)


@st.composite
def hypo_shape(draw, max=1000):
    s = []
    for i in range(3):
        s.append(draw(st.integers(min_value=1, max_value=int(max))))
        max /= s[-1]
    return tuple(s)


@st.composite
def hypo_vars(draw, skip=[]):
    files = []
    for v in load.files:
        if v in (
            "BFIELD_STRENGTH",
            "fort.70",
            "ADD_SF_N0",
            "TARGET_PROFILES",
        ):
            continue
        if v in skip:
            continue
        if draw(st.booleans()):
            t = [v]
            for k in load.files[v]["vars"]:
                if "%" in k:
                    t.append(draw(st.integers(min_value=1, max_value=7)))
            files.append(t)
    return files


@st.composite
def hypo_vars12(draw, skip_info=False):
    skip = []
    if skip_info:
        skip += [
            "NEUTRAL_INFO",
            "ENERGY_INFO",
            "IMPURITY_INFO",
            "STREAMING_INFO",
            # Merging of INFO files isn't supported yet, we just keep
            # the last one ...
        ]
    files = draw(hypo_vars(skip=skip))
    # Skip some random files
    skip += [
        # Also always skip PLATES_MAG:
        #   * it shouldn't change in a simulation
        #   * it is likely to be the same twice, thus resulting in
        #     a test failure.
        "PLATES_MAG",
        # Also always skip LG_CELL:
        #   * it is likely to be the same twice, thus resulting in
        #     a test failure.
        "LG_CELL",
    ]
    files2 = [f for f in files if draw(st.booleans()) and f[0] not in skip]
    assume(len(files2))
    return files, files2


def gen_ds(shape):
    ds = xr.Dataset()

    ds["_plasma_map"] = gen_mapping(shape)
    ds.emc3["bf_corners"] = gen_bf(shape)
    ds.emc3["R_corners"] = gen_bf(shape)
    ds.emc3["z_corners"] = gen_bf(shape)
    ds.emc3["phi_corners"] = ("phi",), np.random.random(shape[2] + 1)
    ds["ne"] = gen_mapped(ds)
    ds["ne"].attrs["print_before"] = "   1\n"
    # ds["nZ1"] = gen_mapped(ds)
    ds["Te"] = gen_mapped(ds)
    ds["Ti"] = gen_mapped(ds)
    ds["nH"] = gen_mapped(ds, True)

    return ds


def gen_full(shape, index=None):
    return gen_rand(shape, None)


def gen_rand(shape, files):
    ds = xr.Dataset()

    ds["_plasma_map"] = gen_mapping(shape)
    ds.emc3["bf_corners"] = gen_bf(shape)
    ds["bf_bounds"].attrs = {"units": "T", "long_name": "Magnetic field strength"}
    ds.emc3["R_corners"] = gen_bf(shape)
    ds.emc3["z_corners"] = gen_bf(shape)
    for k in "R_bounds", "z_bounds":
        ds[k].attrs["units"] = "m"
    ds.emc3["phi_corners"] = ("phi",), np.random.random(shape[2] + 1)
    ds["phi_bounds"].attrs["units"] = "radian"

    def get_attrs(vsv):
        out = {}
        if "attrs" in vsv:
            out.update(vsv["attrs"])
        for attr in "long_name", "units", "notes":
            if attr in vsv:
                out[attr] = vs[v][attr]
        return out

    for f in load.files:
        if f in ["ADD_SF_N0", "TARGET_PROFILES"]:
            continue
        ids = 3
        if files is not None:
            for f2 in files:
                if f2[0] == f:
                    try:
                        ids = f2[1]
                    except IndexError:
                        ids = 3
                    break
            else:
                continue
        if f in ("PARTICLE_DEPO", "ENERGY_DEPO"):
            ids = 6
        i = 0
        vs = load.files[f]["vars"]
        for v in vs:
            if v in ds:
                ds[v].attrs.update(get_attrs(vs[v]))
                continue

            genf = {
                "mapped": gen_mapped,
                "full": gen_bf,
                "plates_mag": gen_plates_mag,
                "info": gen_info,
                "raw": gen_raw,
                "depo": gen_depo,
            }[load.files[f].get("type", "mapped")]
            if load.files[f].get("kinetic", False):
                assert genf == gen_mapped
                genf = gen_kinetic
            pre = load.files[f].get("skip_first", 0)
            dtype = load.files[f].get("dtype", float)

            if "%" in v:
                for i in range(i, i + ids):
                    ds[v % i] = genf(ds, index=i)
                    if dtype != float and genf != gen_depo:
                        ds[v % i] = genf(ds)[0], np.round(genf(ds)[1] * 20)
                    ds[v % i].attrs.update(get_attrs(vs[v]))
                    if pre:
                        ds[v % i].attrs["print_before"] = "   %d\n" % i
            else:
                ds[v] = genf(ds, index=i)
                ds[v].attrs.update(get_attrs(vs[v]))
                if pre:
                    ds[v].attrs["print_before"] = "   %d\n" % i
                if genf == gen_depo and i == 0:
                    ds[v].attrs["description"] = "True means +1, False means -1"

            i += 1
    return ds


def gen_bf(shape, index=None):
    if isinstance(shape, xr.Dataset):
        shape = shape["_plasma_map"].data.shape
    return dims, 0.5 + 2 * np.random.random([i + 1 for i in shape])


def gen_plates_mag(shape, index=None):
    if isinstance(shape, xr.Dataset):
        shape = shape["_plasma_map"].data.shape
    return dims, (0.5 + 1 * np.random.random(shape) > 1)


def gen_kinetic(ds, index=None):
    return gen_mapped(ds, True)


def gen_mapped(ds, kinetic=False, index=None):
    key = "other" if kinetic else "plasmacells"
    map = ds["_plasma_map"]
    shape = map.shape
    max = map.attrs[key]
    dat = np.random.random(max)
    ret = np.zeros(shape) * np.nan
    mapdat = map.values
    for ijk in utils.rrange(shape):
        i = mapdat[ijk]
        if i < max:
            ret[ijk] = dat[i]
    return map.dims, ret


def gen_updated(org: xr.Dataset, var) -> xr.Dataset:
    shape = org["_plasma_map"].shape
    dt = gen_rand(shape, var)
    d2 = org.copy()
    for k in dt:
        if k in (
            "_plasma_map",
            "bf_bounds",
            "R_bounds",
            "z_bounds",
            "phi_bounds",
        ):
            continue
        d2[k] = dt[k]
    return d2


def gen_mapping(shape, index=None):
    dat = np.zeros(shape, dtype=int)
    i = 0
    for ijk in utils.rrange(shape):
        dat[ijk] = i // 3
        i += 1
    da = xr.DataArray(dat, dims=dims)
    pc = i // 6
    if pc == 0:
        pc = 1
    da.attrs = dict(numcells=i, plasmacells=pc, other=np.max(dat) + 1)
    return da


def gen_info(ds: xr.Dataset, index=None) -> xr.DataArray:
    index = "iteration"
    length = np.random.randint(2, 6)
    dat = np.empty(1000)
    dat[:] = np.nan
    dat[-length:] = np.random.random(length)
    coords: typing.Mapping[typing.Hashable, typing.Any] = {index: np.arange(-999, 1)}
    return xr.DataArray(dat, dims=index, coords=coords)


def gen_raw(ds: xr.Dataset, index=None) -> xr.DataArray:
    text = """Some text.
Sooo random.
With linebreaks 🎉
"""
    return xr.DataArray(text)


gen_depo_map = None


def gen_depo(ds: xr.Dataset, index=None):
    shape = ds["_plasma_map"].data.shape
    shape = [x + 1 for x in shape]
    dtype = bool if index in (0, 6) else float
    out = np.empty(shape, dtype=dtype)
    fill = np.nan
    if index == 0:
        fill = True
    if index == 6:
        fill = False
    out[:] = fill

    global gen_depo_map

    if index == 0 or index == 6:
        gen_depo_map = np.random.random(shape) > 0.8

    rand = np.random.random(shape)[gen_depo_map]
    if index == 0:
        out[gen_depo_map] = rand > 0.5
    elif index == 6:
        out[gen_depo_map] = True
    else:
        out[gen_depo_map] = rand

    return tuple([f"{x}_plus1" for x in ds["_plasma_map"].dims]), out


class rotating_circle(object):
    def __init__(self, period, iota=1, sym=False):
        self.period = period
        self.iota = iota
        self.sym = sym
        self.R = 5
        self.r = 1

    def gends(self, shape):
        phi = np.linspace(
            0, np.pi * (1 if self.sym else 2) / self.period, shape[2] + 1
        )  # [None, None, :]
        theta = np.linspace(0, np.pi * 2, shape[1] + 1)[None, :, None] - (
            np.pi / shape[1]
        )

        r = np.linspace(0, self.r, shape[0] + 1)[:, None, None]
        r, p, z = self.rpt_to_rpz(r, phi, theta)

        ds = xr.Dataset()
        ds["_plasma_map"] = gen_mapping(shape)
        ds.emc3["bf_corners"] = gen_bf(shape)
        ds.emc3["R_corners"] = dims, r
        ds.emc3["z_corners"] = dims, z
        ds.emc3["phi_corners"] = ("phi",), p
        return ds

    def rpt_to_rpz(self, r, p, t):
        angle = t + self.iota * p
        r_corner = self.R + r * np.cos(angle)
        z_corner = r * np.sin(angle)
        return r_corner, p, z_corner

    def rpz_to_xyz(self, r, p, z):
        x = r * np.cos(p)
        y = r * np.sin(p)
        return x, y, z


if __name__ == "__main__":
    print(load.files["LG_CELL"])
    ds = gen_full((3, 4, 5))
    print(load.files["LG_CELL"])
    print(ds)



import os
from argparse import ArgumentParser

from .. import __version__


def commonparser(desc, ext="nc") -> ArgumentParser:
    parser = ArgumentParser(description=desc)
    parser.add_argument(
        "path",
        nargs="+",
        help=f"""Path of the directory to load. The netcdf file will be
        called dir.{ext} if the folder was called dir.""",
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="Be less verbose")
    parser.add_argument(
        "-o",
        "--name",
        help=f"Specify the name for the output file. Defaults to `dir.{ext}` "
        f"when not given. Otherwise `name.{ext}` is used.",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"xemc3 {__version__}"
    )
    return parser


def iter_dir(args):
    if len(args.path) > 1 and args.name:
        raise RuntimeError(
            """Providing an explicit output name and more than one folder to be read is
currently not supported. Either use the default output name or read one file
after another. Further you can use the python interface that gives more
control."""
        )
    for d in args.path:
        while d[-1] == "/":
            d = d[:-1]
        if args.name:
            yield d, args.name
        else:
            if os.path.isdir(d):
                if d == ".":
                    yield d, os.getcwd()
                else:
                    yield d, d
            else:
                yield d, d.rsplit(".", 1)[0]



#!/usr/bin/env python3


print(
    """CLI reference
=============

"""
)

path = __file__.split("/")
if len(path) < 2:
    path = ["."]
else:
    path = path[:-1]
path += ["..", "setup.cfg"]
rec = False
found = []
with open("/".join(path)) as f:
    for l in f:
        if rec:
            if l:
                found.append(l)
            else:
                rec = False
        if l.startswith("console_scripts"):
            rec = True

for l in found:
    name, ref = [x.strip() for x in l.split("=")]
    ref = ref.split(":")[0]
    print(
        f"""

.. sphinx_argparse_cli::
   :module: {ref}
   :func: parser
   :prog: {name}
"""
    )



import hypothesis.strategies as strat
import numpy as np
import xarray as xr
from hypothesis import assume, given

from xemc3.core import utils


@given(strat.lists(strat.integers(min_value=1, max_value=10000)))
def test_rrange(shape):
    assume(utils.prod(shape) < 1e5)
    dat = np.zeros(shape, dtype=int)
    for ijk in utils.rrange(shape):
        dat[ijk] += 1
    assert (dat == 1).all()


@given(strat.lists(strat.integers(min_value=1, max_value=100), min_size=0))
def test_rrange2(shape):
    assume(utils.prod(shape) < 1e4)
    dat = np.zeros(shape, dtype=int)
    print(dat)
    for ijk in utils.rrange2(shape):
        dat[ijk] += 1
    assert (dat == 1).all()


def test_merge_indexers_raise():
    a = "a"
    b = "b"
    v1 = {a: 1, b: 2}
    v2 = {a: 2}
    try:
        utils.merge_indexers(v1, v2)
    except AssertionError:
        pass
    else:
        raise AssertionError


def test_merge_indexers_posion():
    a = "a"
    b = "b"
    v1 = {a: 1, b: 2}

    assert v1 == utils.merge_indexers(v1, {})
    assert v1 == utils.merge_indexers(None, v1)


def test_merge_indexers_combine():
    a = "a"
    b = "b"
    v1 = {a: 1}
    v2 = {b: 2}
    both = {**v1, **v2}
    assert both == utils.merge_indexers(v1, v2)
    assert both == utils.merge_indexers(v2, v1)


def test_interval_round_1d():
    a = xr.DataArray(np.random.random(5), dims="a")
    b = utils.from_interval(utils.to_interval(a))
    assert all(a == b)


def test_interval_round_mismatch():
    a = xr.DataArray(np.random.random(5), dims="a")
    b = utils.to_interval(a)
    b.data[1, 1] += 1
    raised = False
    try:
        utils.from_interval(b)
    except:
        raised = True
    assert raised


def test_timit():
    with utils.timeit():
        pass

    with utils.timeit("pass takes %f seconds"):
        pass


def test_timit2():
    with utils.timeit2():
        pass



import copy
import os
import runpy
import sys
import tempfile
import warnings

import xarray as xr
from hypothesis import given, settings
from hypothesis import strategies as st

from .. import write
from . import gen_ds as g
from .test_write_load import assert_ds_are_equal


def call(cmd: str) -> None:
    _argv = copy.deepcopy(sys.argv)
    args = cmd.split()
    sys.argv = args
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            "'xemc3.cli.append_time' found in sys.modules after import of package 'xemc3.cli', but prior to execut",
        )
        runpy.run_module("xemc3.cli." + args[0], run_name="__main__")
    sys.argv = _argv


@settings(**g.setting)  # type: ignore
@given(
    g.hypo_shape(10),
    g.hypo_vars12(),
    st.integers(min_value=1, max_value=3),
)
def test_append_ds(shape, v12, rep):
    do_test_append_ds(shape, v12, rep)


def do_test_append_ds(shape, v12, rep):
    v1, v2 = v12
    assert len(v2)
    orgs = [g.gen_rand(shape, v1)]
    out = "temp"
    for i in range(rep):
        orgs.append(g.gen_updated(orgs[0], v2))
    with tempfile.TemporaryDirectory() as dir:
        dir = dir + "/" + out
        os.mkdir(dir)
        for org in orgs:
            write.fortran(org, dir)
            call("append_time " + dir + "///")
        nc = dir + ".nc"
        with xr.open_dataset(nc) as read:
            if "time" not in read.dims:
                assert_ds_are_equal(orgs[0], read, True, 1e-2, 1e-2)
            else:
                for i, org in enumerate(orgs):
                    assert_ds_are_equal(org, read.isel(time=i), True, 1e-2, 1e-2)


@settings(**g.setting)  # type: ignore
@given(
    g.hypo_shape(30),
    g.hypo_vars(),
    st.integers(min_value=1, max_value=3),
)
def test_to_netcdf_ds(shape, var, rep):
    do_to_netcdf_ds(shape, var, rep)


def do_to_netcdf_ds(shape, var, rep):
    out = "test"
    with tempfile.TemporaryDirectory() as dir:
        dir = dir + "/" + out
        os.mkdir(dir)
        for i in range(rep):
            org = g.gen_rand(shape, var)
            write.fortran(org, dir)
            call("to_netcdf " + dir + "///")
            nc = dir + ".nc"
            read = xr.open_dataset(nc)
            assert_ds_are_equal(org, read, True, 1e-2, 1e-2)
            del read
        os.remove(nc)


if __name__ == "__main__":
    do_test_append_ds((2, 2, 2), ([["ENERGY_INFO"]], [["ENERGY_INFO"]]), 1)
    # do_to_netcdf_ds((1, 1, 1), [["ENERGY_INFO"]], 1)



import os

import xarray as xr


def download_file():
    os.system(
        "git clone https://oauth2:uQ7_vw_m5fo4UFQrenUs@gitlab.mpcdf.mpg.de/dave/xemc3-data/ ../../example-data/ --depth 1"
    )


def load_example_data(get_path=False):
    # Check whether we have a local copy
    for path in "./", "../", "../../":
        for path in [path, path + "example-data/"]:
            try:
                fn = path + "emc3_example.nc"
                if get_path:
                    xr.open_dataset(fn)
                    return fn[:-3]
                return xr.open_dataset(fn)
            except FileNotFoundError:
                pass
            except ValueError:
                print("Skipping non-readable file", fn)

    # try file from AFS
    try:
        ds = xr.open_dataset(
            "/afs/ipp-garching.mpg.de/u/dave/public/xemc3/emc3_example.nc"
        )
        print("Accessing example file via AFS, which might be slow ...")
        return ds
    except:  # noqa: E722
        # Might fail for a number of reasons ...
        pass

    # Last resort, download and use that
    print("Trying to download an example file to ../../example-data/ ...")
    download_file()
    print("done")
    return xr.open_dataset("../../example-data/emc3_example.nc")



import datetime
import os
import re
import typing
import uuid

import numpy as np
import xarray as xr

from .utils import from_interval, open, prod, rrange, timeit, to_interval
from .depo import read_depo_raw, write_depo_raw

try:
    from numba import jit  # type: ignore
except ImportError:

    def jit(x):
        return x


try:
    from numpy.typing import DTypeLike
except ImportError:
    # Workaround for python 3.6
    DTypeLike = typing.Type  # type: ignore


def _fromfile(
    f: typing.TextIO, *, count: int, dtype: DTypeLike, **kwargs
) -> np.ndarray:
    """
    Read count amount of dtype from the textio stream, i.e. a file
    opened for reading. It always reads full lines. If there is more
    data on the line then it should read, all data is returned. If
    less data is read, then a shorter array is returned. Thus the
    calling function needs to handle the case that more or less data
    is returned.

    Parameters
    ----------
    f : textio
        the opened file
    count : int
        amount that should be read
    dtype : dtype
        the datatype to expect
    kwargs : dict
        Additional arguments for np.fromstring

    Returns
    -------
    np.ndarray
        The read data.
    """
    ret = np.empty(count, dtype=dtype)  # type: np.ndarray
    pos = 0
    bad = re.compile(r"(\d)([+-]\d)")

    while True:
        line = f.readline()
        if line == "":
            break
        line = bad.sub(r"\1E\2", line)
        new = np.fromstring(line, dtype=dtype, **kwargs)
        ln = len(new)
        try:
            ret[pos : pos + ln] = new
        except Exception as e:
            print(e)
            if pos + ln >= count:
                ret = np.append(ret[:pos], new)
                print(f"Returning {pos + ln} rather then {count} elements")
            else:
                raise
        pos += ln
        if ln == 0 or count <= pos:
            break
    if count != pos:
        ret = ret[:pos]
        return ret
    return ret


def _block_write(f: typing.TextIO, d: np.ndarray, fmt: str, bs: int = 10) -> None:
    d = d.flatten()
    asblock = (len(d) // bs) * bs
    np.savetxt(f, d[:asblock].reshape(-1, bs), fmt=fmt)
    if asblock != len(d):
        np.savetxt(f, d[asblock:], fmt=fmt)


def write_locations(ds: xr.Dataset, fn: str) -> None:
    """
    Write spatial positions of grid points in EMC3 format.

    Parameters
    ----------
    ds : xr.Dataset
        The dataset with the EMC3 simulation data
    fn : str
        The filename to write to
    """
    rs = ds.emc3["R_corners"]
    phis = ds.emc3["phi_corners"]
    zs = ds.emc3["z_corners"]
    fmt = "%10.6f"

    def write(f, d):
        d = d.transpose()
        _block_write(f, d, fmt)

    with open(fn, "w") as f:
        # np.savetxt(f, t, "%d")
        f.write("  ".join([str(t) for t in rs.shape]) + "\n")
        for i, phi in enumerate(phis):
            np.savetxt(f, [phi * 180 / np.pi], fmt=fmt)
            write(f, rs.isel(phi=i).data * 100)
            write(f, zs.isel(phi=i).data * 100)


def read_magnetic_field(fn: str, ds: xr.Dataset) -> xr.DataArray:
    """
    Read magnetic field strength from grid

    Parameters
    ----------
    fn : str
        The filename to read from
    ds : xr.Dataset
        A xemc3 dataset required for the dimensions

    Returns
    -------
    xr.Dataset
        The magnetic field strength
    """
    if "R_bounds" in ds:
        shape = ds.R_bounds.shape
        assert len(shape) == 6
        shape = [i + 1 for i in shape[:3]]
        dims = ds.R_bounds.dims[:3]
    else:
        shape = ds._plasma_map.shape
        shape = [i + 1 for i in shape]
        dims = ds._plasma_map.dims
    nx, ny, nz = shape
    with open(fn) as f:
        raw = _fromfile(f, dtype=float, count=nx * ny * nz, sep=" ")
        _assert_eof(f, fn)
    raw = raw.reshape(shape[::-1])
    raw = np.swapaxes(raw, 0, 2)
    return to_interval(dims, raw)


def write_magnetic_field(path: str, ds: xr.Dataset) -> None:
    """
    Write the magnetic field to a file

    Parameters
    ----------
    path : str
        The filename to write to
    ds : xr.Dataset
        The xemc3 dataset with the magnetic field
    """
    bf = ds.emc3["bf_corners"].data
    bf = bf.transpose(2, 1, 0)
    with open(path, "w") as f:
        _block_write(f, bf, "%7.4f")


def read_locations_raw(fn: str) -> typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Read spatial positions of grid points

    Parameters
    ----------
    fn : str
        The filename of the grid

    Returns
    -------
    np.array
        phi positions
    np.array
        radial positions
    np.array
        z position
    """
    with open(fn) as f:
        nx, ny, nz = [int(i) for i in f.readline().split()]
        phidata = np.empty(nz)
        rdata = np.empty((nx, ny, nz))
        zdata = np.empty((nx, ny, nz))

        def read(f, nx, ny):
            t = _fromfile(f, dtype=float, count=nx * ny, sep=" ")
            t = t.reshape(ny, nx)
            t = t.transpose()
            return t

        # Read and directly convert to SI units
        for i in range(nz):
            phidata[i] = float(f.readline()) * np.pi / 180.0
            rdata[:, :, i] = read(f, nx, ny) / 100
            zdata[:, :, i] = read(f, nx, ny) / 100

        _assert_eof(f, fn)

    return phidata, rdata, zdata


def read_plates_mag(fn: str, ds: xr.Dataset) -> xr.DataArray:
    """
    Read the magnetic plates

    Parameters
    ----------
    fn : str
        Location of the magnetic plates files
    ds : xr.Dataset
        xemc3 dataset

    Returns
    -------
    xr.DataArray
        The magnetic plates
    """
    shape = [x.shape[0] for x in [ds.r, ds.theta, ds.phi]]
    ret = np.zeros(shape, dtype=bool)
    with open(fn) as f:
        last = None
        for i, line in enumerate(f):
            lines = [int(x) for x in line.split()]
            if last:
                lines = last + lines
            zone, r, theta, num = lines[:4]
            assert zone == 0
            assert num % 2 == 0
            if num + 4 > len(lines):
                last = lines
                continue
            assert num + 4 == len(lines), (
                f"failed to parse line {i+1}{' (continued from previous incomplete line)' if last else ''}"
                f" from {fn}: {' '.join([str(x) for x in lines])}"
            )
            last = None
            for t in range(num // 2):
                a, b = lines[4 + 2 * t : 6 + 2 * t]
                ret[r, theta, a : b + 1] = True
    return xr.DataArray(data=ret, dims=("r", "theta", "phi"))


def write_plates_mag(fn: str, ds: xr.Dataset) -> None:
    """
    Write the ``PLATES_MAG`` info to a file in the EMC3 format.

    Parameters
    ----------
    fn : str
        The file to write to
    ds : xr.Dataset or xr.DataArray
       either the PLATES_MAG dataarray or a dataset containing PLATES_MAG
    """
    if isinstance(ds, xr.Dataset):
        da = ds["PLATES_MAG"]
    else:
        da = ds
    with open(fn, "w") as f:
        # iterate over r:
        for r, slice2d in enumerate(da.data):
            if not slice2d.any():
                continue
            for theta, slice1d in enumerate(slice2d):
                if not slice1d.any():
                    continue
                if slice1d.all():
                    ab = [0, len(slice1d)]
                else:
                    last = False
                    ab = []
                    for i, v in enumerate(slice1d):
                        if v != last:
                            if v or i == 0:
                                ab.append(i)
                            else:
                                ab.append(i - 1)
                            last = v
                    if last:
                        ab.append(i)
                out = [0, r, theta, len(ab)] + ab
                outi = [str(i) for i in out]
                outs = "   %s   %3s   %3s" % tuple(out[:3])
                outs += " ".join(["    %2s" % i for i in outi[3:]])
                f.write(outs + "\n")
    return


def read_mappings(fn: str, dims: typing.Sequence[int]) -> xr.DataArray:
    """
    Read the mappings data

    Parameters
    ----------
    fn : str
        The path of the file to be read
    dims : tuple of int
        The shape of the domain

    Returns
    -------
    xr.DataArray
        The mapping information
    """
    with open(fn) as f:
        dat = f.readline()
        infos = [int(i) for i in dat.split()]
        t = _fromfile(f, dtype=int, count=prod(dims), sep=" ")
        # fortran indexing
        t -= 1
        t = t.reshape(dims, order="F")
        _assert_eof(f, fn)
    da = xr.DataArray(dims=("r", "theta", "phi"), data=t)
    da.attrs = dict(numcells=infos[0], plasmacells=infos[1], other=infos[2])
    return da


def add_metadata(ds: xr.Dataset):
    # Delay import to avoid circular dependency
    from .. import __version__

    ds.attrs.update(
        title="EMC3-EIRENE Simulation data",
        software_name="xemc3",
        software_version=__version__,
        date_created=datetime.datetime.utcnow().isoformat(),
        id=str(uuid.uuid1()),
        references="https://doi.org/10.5281/zenodo.5562265",
    )


def write_mappings(da: xr.DataArray, fn: str) -> None:
    """Write the mappings data to fortran"""
    with open(fn, "w") as f:
        at_keys = "numcells", "plasmacells", "other"
        for key in at_keys:
            assert key in da.attrs, (
                "Writing requires to be the following attributes to be present:"
                + ", ".join([f'"{k}"' for k in at_keys])
                + ". Please ensure that they are copied over from the read file."
            )
        infos = tuple([da.attrs[k] for k in at_keys])
        f.write("%12d %11d %11d\n" % infos)
        _block_write(f, da.data.flatten(order="F") + 1, " %11d", 6)


def read_raw(fn: str) -> xr.DataArray:
    with open(fn) as f:
        return xr.DataArray(f.read())


def write_raw(da: xr.DataArray, fn: str) -> None:
    with open(fn, "w") as f:
        f.write(str(da.data))


def read_locations(path: str, ds: xr.Dataset = None) -> xr.Dataset:
    """
    Read locations from folder path and add to dataset (if given).

    Parameters
    ----------
    path : str
        folder from which to read
    ds : xr.Dataset (optional)
        The dataset to add to. If not given return a new dataset.

    Returns
    -------
    xr.Dataset
        A dataset with the coordinates set
    """
    if ds is None:
        ds = xr.Dataset()
    assert isinstance(ds, xr.Dataset)
    phi, r, z = read_locations_raw(path + "/GRID_3D_DATA")
    ds = ds.assign_coords(
        {
            "R_bounds": to_interval(("r", "theta", "phi"), r),
            "z_bounds": to_interval(("r", "theta", "phi"), z),
            "phi_bounds": to_interval(("phi",), phi),
        }
    )
    for x in ds.coords:
        ds[x].attrs["xemc3_type"] = "geom"
    ds.emc3.unit("R_bounds", "m")
    ds.emc3.unit("z_bounds", "m")
    ds.emc3.unit("phi_bounds", "radian")
    assert isinstance(ds, xr.Dataset)
    add_metadata(ds)
    return ds


def scrape(f: typing.TextIO, *, ignore="!", verbose=False) -> str:
    """read next data line from configuration file (skip lines with
    leading *)

    Parameters
    ----------
    f : textio
        fileobject (opened file) to read from
    ignore : (str or None)
        if any of the characters are given, interpret as comment and
        remove. In this case only return non-empty lines.

    Returns
    ------
    str
        the first non-ignored line
    """
    while True:
        s = f.readline()
        if s.startswith("*"):
            if verbose:
                if s.startswith("***"):
                    print(s[:-1])
            continue
        if ignore is not None:
            for i in ignore:
                if i in s:
                    s = s[: s.index(i)]
            if s.strip() == "":
                continue
        return s


def _assert_eof(f: typing.TextIO, fn: str) -> None:
    """Ensure the opened file is read to the end.

    Parameters
    ----------
    f : textio
        the opened file
    fn : str
        The name of the file, used for the error.

    Raises
    ------
    AssertionError
        if the file is not at the end.
    """
    test = _fromfile(f, dtype=float, count=2, sep=" ")
    assert len(test) == 0, f"Expected EOF, but found more data in {fn}"


def read_plate(filename: str) -> typing.Tuple[np.ndarray, ...]:
    """
    Read Target structures from a file that is in the Kisslinger
    format as used by EMC3. It returns the coordinates as plain array
    in the order or R, z, phi.

    Parameters
    ----------
    filename : str
        The location of the file to read

    Returns
    -------
    np.ndarray
        The major radius coordinates of the corners in meters.
    np.ndarray
        The z coordinate of the corners in meters
    np.ndarray
        the phi coordinate of the corners in radian
    """
    with open(filename) as f:
        # first line is a comment ...
        _ = next(f)
        setup = next(f).split()
        assert len(setup) == 5, f"Expected 5 values but got {setup}"
        for zero in setup[3:]:
            assert float(zero) == 0.0
        nx, ny, nz = [int(i) for i in setup[:3]]
        r = np.empty((nx, ny))
        z = np.empty((nx, ny))
        phi = np.empty(nx)
        for x in range(nx):
            phi[x] = float(next(f)) / 180 * np.pi
            for y in range(ny):
                s = next(f)
                try:
                    try:
                        r[x, y], z[x, y] = [float(i) / 100 for i in s.split()]
                    except ValueError:
                        s = s.split("!")[0]
                        r[x, y], z[x, y] = [float(i) / 100 for i in s.split()]
                except ValueError:
                    raise ValueError(f"Error with {s} in {filename}")
        _assert_eof(f, filename)
        return (r, z, phi)


def read_plate_nice(filename: typing.Union[str, typing.Sequence[str]]) -> xr.Dataset:
    """
    Read Target structures from a file that is in the Kisslinger
    format as used by EMC3.

    Parameters
    ----------
    filename : str or sequence of str
        The location of the file to read

    Returns
    -------
    xr.Dataset
        The coordinates
    """
    if isinstance(filename, str):
        return read_plate_ds(filename)
    dss = [read_plate_ds(fn) for fn in filename]
    return merge_blocks(dss)


def read_add_sf_n0(filename: str) -> xr.Dataset:
    """
    Read the ADD_SF_N0 file and return the referenced geometry.

    Parameters
    ----------
    filename : str
        The path of the file

    Returns
    -------
    xr.Dataset
        The coordinates
    """
    if "/" in filename:
        dir = "/".join(filename.split("/")[:-1]) + "/"
    else:
        dir = ""
    files = []
    with open(filename) as f:
        num = int(scrape(f))
        for _ in range(num):
            line = scrape(f)
            lines = line.split()
            assert len(lines) == 3, f"Unexpected content in {filename}: {line}"
            assert [int(x) for x in lines] == [
                0,
                -4,
                1,
            ], f"Unexpected content in {filename}: {line}"
            files.append(dir + scrape(f).strip())
    return read_plate_nice(files)


plate_prefix = "plate_"


def read_plate_ds(filename: str) -> xr.Dataset:
    """
    Read Target structures from a file that is in the Kisslinger
    format as used by EMC3. It returns the coordinates as dataset
    containing R, z and phi.

    Parameters
    ----------
    filename : str
        The location of the file to read

    Returns
    -------
    xr.Dataset
        The coordinates
    """
    rzp = read_plate(filename)
    out = xr.Dataset()
    for name, dat in zip(("R", "z", "phi"), rzp):
        if len(dat.shape) == 2:
            dims = ["phi", "x"]
        else:
            assert len(dat.shape) == 1
            dims = ["phi"]
        out = out.assign_coords(
            {f"{plate_prefix}{name}": ([f"{plate_prefix}{d}_plus1" for d in dims], dat)}
        )
    for var, attrs in files["ADD_SF_N0"]["vars"].items():
        out[var].attrs.update(attrs)

    return out


plates_labels = ["f_n", "f_E", "avg_n", "avg_Te", "avg_Ti"]


def read_plates_raw(cwd: str, fn: str) -> typing.Sequence[xr.Dataset]:
    """
    Read the target mapped info to a list of xr.Datasets

    Parameters
    ----------
    cwd : str
        The folder in which to look for the files (with trailing slash)
    fn : str
        the relative name of the file to read

    Returns
    -------
    list of xr.Dataset
        The read data as a list of datasets
    """
    with open(cwd + fn) as f:
        s = scrape(f).split()
        _, num_plates = [int(i) for i in s]
        plates = []
        for plate in range(num_plates):
            s = scrape(f).split()
            assert len(s) == 2, f"{plate} : {s}"
            _, geom = s
            r, z, phi = read_plate(cwd + geom)
            nx, ny = r.shape
            nx -= 1
            ny -= 1
            s = scrape(f).split()
            total = np.array([float(i) for i in s])
            s = scrape(f).split()
            if len(s) == 3:
                items, yref, xref = [int(i) for i in s]
                nxr = nx * xref
                nyr = ny * yref
                mode = 2
            else:
                assert len(s) == 2, f"Unexpected string `{s}` while reading {cwd + fn}"
                nyr, nxr = [int(i) for i in s]
                xref = nxr // nx
                yref = nyr // ny
                items = nx * ny
                mode = 1

            assert items == nx * ny
            if mode == 2:
                data = _fromfile(f, dtype=float, count=nx * ny * 12, sep=" ")

                corrs = [data[4 * items * i : 4 * items * (i + 1)] for i in range(3)]
                corrs[0] /= 100
                corrs[1] /= 100
                # phi is in rads ?!?
                # corrs[2] /= 180/np.pi
                corrs = [a.reshape(nx, ny, 4) for a in corrs]
            else:
                assert mode == 1
                data = _fromfile(
                    f, dtype=float, count=(nxr + 1) * (nyr + 1) * 3, sep=" "
                )

                coordinates = [
                    data[(nxr + 1) * (nyr + 1) * i : (nxr + 1) * (nyr + 1) * (i + 1)]
                    for i in range(3)
                ]
                coordinates[0] /= 100
                coordinates[1] /= 100
                coordinates = [a.reshape(nxr + 1, nyr + 1) for a in coordinates]

            data = _fromfile(f, dtype=float, count=nxr * nyr * 5, sep=" ")
            if mode == 2:
                data = data.reshape((5, nx, ny, xref, yref))
                data = data.transpose((0, 1, 3, 2, 4))
            data = data.reshape(5, nxr, nyr)
            if phi[0] > phi[1]:
                phi = phi[::-1]
                r = r[::-1, ...]
                z = z[::-1, ...]
                data = data[:, ::-1, ...]
                if mode == 2:
                    corrs = [a[::-1, ...] for a in corrs]
                else:
                    coordinates = [a[::-1, ...] for a in coordinates]

            if mode == 2:
                A = np.array(
                    [(1.0, a, b, a * b) for a, b in [(0, 0), (0, 1), (1, 1), (1, 0)]]
                ).T
                Ai = np.linalg.inv(A)
                a0, a1 = 0, 1
                b0, b1 = 0, 1
                A = np.array(
                    [
                        (1.0, a, b, a * b)
                        for a, b in [(a0, b0), (a0, b1), (a1, b0), (a1, b1)]
                    ]
                ).T
                for i, pos in enumerate(corrs):
                    # Transform into coefficient form
                    pos = pos @ Ai
                    newpos = np.empty((nxr, nyr, 4))
                    for ix, iy in rrange((xref, yref)):
                        # Get relative coordinates
                        a0, a1 = ix / xref, (ix + 1) / xref
                        b0, b1 = iy / yref, (iy + 1) / yref
                        A = np.array(
                            [
                                (1.0, b, a, a * b)
                                for a, b in [(a0, b0), (a0, b1), (a1, b0), (a1, b1)]
                            ]
                        ).T
                        newpos[ix::xref, iy::yref, :] = pos @ A

                    corrs[i] = newpos
                corrs_da = [
                    xr.DataArray(
                        data=a.reshape(nxr, nyr, 2, 2),
                        dims=("phi", "x", "delta_phi", "delta_x"),
                    )
                    for a in corrs
                ]
            else:
                assert mode == 1
                corrs_da = [
                    to_interval((plate_prefix + "phi", plate_prefix + "x"), a)
                    for a in coordinates
                ]

            coords = {
                plate_prefix + "R_bounds": corrs_da[0],
                plate_prefix + "z_bounds": corrs_da[1],
                plate_prefix + "phi_bounds": corrs_da[2],
            }
            ds = xr.Dataset(coords=coords)  # type: ignore
            ds.coords[plate_prefix + "R_bounds"].attrs["units"] = "m"
            ds.coords[plate_prefix + "z_bounds"].attrs["units"] = "m"
            ds.coords[plate_prefix + "phi_bounds"].attrs["units"] = "radian"

            vars = files["TARGET_PROFILES"]["vars"].copy()
            for i, (l, meta) in enumerate(vars.items()):
                ds[l] = (plate_prefix + "phi", plate_prefix + "x"), data[i] * meta.get(
                    "scale", 1
                )
                for k in meta:
                    ds[l].attrs[k] = meta[k]
            for i, l in enumerate(["tot_n", "tot_P"]):
                ds[l] = total[i]
            plates.append(ds)

        # Make sure we have read everything
        _assert_eof(f, cwd + fn)
    return plates


def merge_blocks(
    dss: typing.Sequence[xr.Dataset], axes=plate_prefix + "ind"
) -> xr.Dataset:
    """
    Convert a list of datasets to one dataset

    Parameters
    ----------
    dss : sequence of xr.Datasets
        The sequence of datasets that should be merged. Unlike
        xr.combine_nested the input's do not need to have the same
        shape, as the data is nan-padded.

    axis : str
        The name of the new axis

    Returns
    -------
    xr.Dataset
        The merged dataset with the new axes
    """
    dims = {d: 0 for d in dss[0].dims}
    for plate in dss:
        for k, v in plate.dims.items():
            if dims[k] < v:
                dims[k] = v
    ds = xr.Dataset()

    matching = {}
    for k, v in dims.items():
        matching[k] = True
        org_dims = []
        for plate in dss:
            org_dims.append(plate.dims[k])
            if plate.dims[k] != v:
                matching[k] = False
        if not matching[k]:
            assert isinstance(k, str)
            ds[f"_{k}_dims"] = (axes, org_dims)

    dims[axes] = len(dss)

    def merge(var):
        shape = [dims[axes]] + [dims[d] for d in dss[0][var].dims]
        data = np.empty(shape)
        data[...] = np.nan
        for i, plate in enumerate(dss):
            tmp = plate[var]
            data[tuple([i] + [slice(None, i) for i in tmp.shape])] = tmp
        return (axes, *dss[0][var].dims), data

    for coord in dss[0].coords.keys():
        ds = ds.assign_coords({coord: merge(coord)})

    for var in dss[0]:
        ds[var] = merge(var)
        ds[var].attrs = dss[0][var].attrs

    return ds


def load_plates(dir: str, fn: str = "TARGET_PROFILES") -> xr.Dataset:
    """
    Read the target heatflux mapping from EMC3 Postprocessing routine.

    Parameters
    ----------
    dir : str
        The location of the directory in which the files are to be read

    fn : str
        The name of the deposition file

    Returns
    -------
    xr.Dataset
        The read data
    """
    if dir[-1] != "/":
        dir += "/"
    plates = read_plates_raw(dir, fn)
    return merge_blocks(plates)


def write_plates(dir: str, plates: xr.Dataset) -> None:
    # Deprecate?
    with timeit("Writing ncs: %f"):
        # Note we really should compress to get rid of the NaN's we added
        plates.to_netcdf(
            f"{dir}/TARGET_PROFILES.nc",
            encoding={
                i: {"zlib": True, "complevel": 1}
                for i in [i for i in plates] + [i for i in plates.coords]
            },
        )


def read_plates(dir: str) -> xr.Dataset:
    # Deprecate?
    ds = xr.open_dataset(f"{dir}/TARGET_PROFILES.nc")
    assert isinstance(ds, xr.Dataset)
    return ds


def _dir_of(fn: str) -> str:
    if os.path.isdir(fn):
        return fn
    if "/" in fn:
        return fn.rsplit("/", 1)[0]
    return "."


def get_plates(dir: str, cache: bool = True) -> xr.Dataset:
    """
    Read the target fluxes from the EMC3_post processing routine

    Parameters
    ----------
    dir : str
        The directory in which the file is
    cache : bool (optional)
        Whether to check whether a netcdf file is present, and if not
        create one. This can result in faster reading the next time.

    Returns
    -------
    xr.Dataset
        The target profiles
    """
    dir = _dir_of(dir)

    if cache:
        try:
            if os.path.getmtime(dir + "/TARGET_PROFILES.nc") > os.path.getmtime(
                dir + "/TARGET_PROFILES"
            ):
                return read_plates(dir)
        except OSError:
            pass
        data = load_plates(dir)
        write_plates(dir, data)
        return data
    return load_plates(dir)


def ensure_mapping(
    dir: str,
    mapping: typing.Union[None, xr.Dataset, xr.DataArray] = None,
    need_mapping: bool = True,
    fn: typing.Union[None, str] = None,
) -> xr.Dataset:
    """
    Ensure that basic info's are present to read datafiles.

    Parameters
    ----------
    dir : str
        The folder in which to look for files
    mapping : None or xr.Dataset or xr.DataArray
        Either the required mapping, in which case nothing is done, or
        None, in which case the data is read.
    need_mapping : bool (optional)
        If true the mapping information has to be loaded, otherwise the
        shape of the data is sufficient. Defaults to True.
    fn : str (optional)
        The file to read, used for a potential error message

    Returns
    -------
    xr.Dataset
        A dataset with the required information

    Raises
    ------
    FileNotFoundError
        If the required info could not be read
    """
    if dir == "":
        dir = "."
    if mapping is None:
        try:
            mapping = read_locations(dir)
            if need_mapping:
                mapping = read_fort_file(mapping, f"{dir}/fort.70", **files["fort.70"])
        except FileNotFoundError:
            raise FileNotFoundError(
                f"""Reading {fn+ ' ' if fn is not None else ''} mapped requires mapping information, but the required
information in '{dir}' could not be found.  Ensure all files are present
in the folder or pass in a dataset that contains the mapping
information."""
            )
    else:
        if isinstance(mapping, xr.DataArray):
            mapping = xr.Dataset(dict(_plasma_map=mapping))
    return mapping


def read_var(
    dir: str, var: str, ds: typing.Union[None, xr.Dataset, xr.DataArray] = None
) -> xr.Dataset:
    """
    Read the variable var from the simulation in directory dir.

    Parameters
    ----------
    dir: str
        The path of the simulation directory
    var: str
        The name of the variable to be read
    ds: xr.DataArray or xr.Dataset or None (optional)
        An xemc3 Dataset or an DataArray containing the mapping. If
        not given and a mapping is required to read the file, it is
        read from the directory.

    Returns
    -------
    xr.Dataset
        A dataset in which at least ``var`` is set. If there are other
        variables in the read file, it is also added. Finally, mapping
        and other variables that are required to read are also added.
    """
    for fn, fd in files.items():
        if "vars" not in fd:
            continue
        for v in fd["vars"]:
            if v == var or (
                v.endswith("%d")
                and var.startswith(v[:-2])
                and var[len(v[:-2]) :].isdigit()
            ):
                ds = ensure_mapping(dir, ds, fd.get("type", "mapped") == "mapped")
                assert isinstance(ds, xr.Dataset)
                return read_fort_file(ds, f"{dir}/{fn}", **fd)
    raise ValueError(f"Don't know how to read {var}")


def read_mapped(
    fn: str,
    mapping: typing.Union[xr.Dataset, xr.DataArray],
    skip_first: int = 0,
    ignore_broken: bool = False,
    kinetic: bool = False,
    dtype: DTypeLike = float,
    squeeze: bool = True,
) -> typing.Sequence[xr.DataArray]:
    """
    Read a file with the EMC3 mapping. Note that this function does
    not add meta-data or does normalisation of the data, thus
    `xemc3.load.file` is generally preferred.

    Parameters
    ----------
    fn : str
        The full path of the file to read
    mapping : xr.DataArray or xr.Dataset
        The dataarray of the mesh mapping or a dataset containing the
        mapping
    skip_first : int (optional)
        Ignore the first n lines. Default 0
    ignore_broken: bool (optional)
        if incomplete datasets at the end should be ignored. Default: False
    kinetic : bool (optional)
        The file contains also data for cells that are only evolved by
        EIRENE, rather then EMC3. Default: False
    dtype : datatype (optional)
        The type of the data, e.g. float or int. Is passed to
        numpy. Default: float
    squeeze : bool
        If True return a DataArray if only one field is read.

    Returns
    -------
    xr.DataArray or list of xr.DataArray
        The data that has been read from the file. If squeeze is True
        and only one field is read only a single DataArray is
        returned.

    """

    if isinstance(mapping, xr.Dataset):
        mapping = ensure_mapping(_dir_of(fn), mapping)
        mapping = mapping["_plasma_map"]
    if kinetic:
        max = np.max(mapping.data) + 1
    else:
        max = mapping.attrs["plasmacells"]
    firsts = []
    with open(fn) as f:
        raws = []
        while True:
            if skip_first:
                first = ""
                for _ in range(skip_first):
                    first += f.readline()
                firsts.append(first)
            raw = _fromfile(f, dtype=dtype, count=max, sep=" ")
            if ignore_broken and len(raw) > max:
                print(f"Ignoring data: {raw[max:]}")
                raw = raw[:max]
            if len(raw) == max:
                raws.append(raw)
            elif len(raw) == 0:
                break
            else:
                if ignore_broken:
                    print(f"Ignoring {len(raw)} data points, {max} required")
                    break
                print(raw)
                raise RuntimeError(
                    f"Incomplete dataset found ({len(raw)} out of {max}) after reading {len(raws)} datasets of file {fn}"
                )

    def to_da(raw):
        out = np.ones(mapping.shape) * np.nan
        mapdat = mapping.data
        for ijk in rrange(mapping.shape):
            mapid = mapdat[ijk]
            if mapid < max:
                out[ijk] = raw[mapid]
        return xr.DataArray(data=out, dims=mapping.dims)

    das = [to_da(raw) for raw in raws]
    if skip_first:
        for first, da in zip(firsts, das):
            da.attrs["print_before"] = first
    if squeeze and len(das) == 1:
        das = das[0]
    return das


def write_mapped_nice(ds: xr.Dataset, dir: str, fn: str = None, **args) -> None:
    """
    Write a file for EMC3 using the mapped format.

    Parameters
    ----------
    ds : xr.Dataset
        A xemc3 dataset
    dir : str
        The directory to which to write the file
    fn : str or None (optional)
        In case of ``None`` all mapped files are written.
        In the case of a str only that file is written.
        Any missing data is ignored. Thus if a file is specified, but
        the data hasn't been loaded this is ignored.
    args : dict (optional)
        Can be used to overwrite options for writing. Defaults to the
        options used for that file.
    """
    meta: typing.Dict[str, typing.Any]
    if fn is None:
        for fn, meta in files.values():  # type: ignore
            assert isinstance(meta, dict)
            if meta["type"] == "mapped":
                write_mapped_nice(ds, dir, fn, **args)
    else:
        meta = files[fn]
        meta = meta.copy()
        meta.update(args)
        meta.pop("type", "ignore")
        ignore_missing = meta.pop("ignore_missing", True)
        try:
            vars = get_vars_for_file(ds, meta.pop("vars"))
        except KeyError:
            if not ignore_missing:
                raise
            return
        datas = [ds[k] for k, _ in vars]
        for i, ops in enumerate([m for _, m in vars]):
            if "scale" in ops:
                at = datas[i].attrs
                datas[i] = datas[i] / ops["scale"]
                datas[i].attrs = at
        assert (
            datas != []
        ), f"Requested to write file {dir}/{fn} but required data not found."
        write_mapped(datas, ds["_plasma_map"], f"{dir}/{fn}", **meta)


def get_vars_for_file(
    ds: xr.Dataset, fn: typing.Union[str, dict]
) -> typing.List[typing.Tuple[str, dict]]:
    """
    Check if the variables of fn are loaded.

    Parameters
    ----------
    ds : xr.Dataset
        An xemc3 dataset
    fn : str or dict
        A file name to check or the associated dictionary of mappings

    Returns
    -------
    list of tuple of str and dict
        The variables that are from that file as well as a dict
        containing additional info about the variable.

    Raises
    ------
    KeyError
        If the file has not been loaded
    """
    vars: dict
    if isinstance(fn, dict):
        vars = fn
    else:
        vars = files[fn]["vars"].copy()
    keys: typing.List[typing.Tuple[str, dict]] = []
    for var in vars:
        # Equivalent writing code:
        # if "%" in keys[-1]:
        #     key = keys[-1]
        #     flexi = vars.pop(key)
        #     for i in range(len(vars), len(datas)):
        #         vars[key % i] = flexi
        if "%" in var:
            while True:
                i = len(keys)
                if var % i not in ds:
                    break
                keys.append((var % i, vars[var]))
        else:
            keys.append((var, vars[var]))
    if keys == []:
        raise KeyError(f"Didn't find any key for {fn}")
    for k, _ in keys:
        if k not in ds:
            raise KeyError(f"Key {k} not present in Dataset, only have {ds.keys()}")
    return keys


@jit
def to_mapped_core(
    datdat: np.ndarray, mapdat: np.ndarray, out: np.ndarray, count: np.ndarray, max: int
) -> typing.Tuple[np.ndarray, np.ndarray]:
    if len(datdat.shape) == 3:
        for i in range(mapdat.shape[0]):
            for j in range(mapdat.shape[1]):
                for k in range(mapdat.shape[2]):
                    mapid = mapdat[i, j, k]
                    if mapid < max:
                        cdat = datdat[(..., i, j, k)]
                        if not (np.isnan((cdat))):
                            out[..., mapid] += cdat
                            count[mapid] += 1
    else:
        for i in range(mapdat.shape[0]):
            for j in range(mapdat.shape[1]):
                for k in range(mapdat.shape[2]):
                    mapid = mapdat[i, j, k]
                    if mapid < max:
                        cdat = datdat[(..., i, j, k)]
                        if not (np.isnan((cdat))):
                            out[..., mapid] += cdat
                            count[mapid] += 1
    return out, count


def to_mapped(
    data: xr.DataArray,
    mapping: xr.DataArray,
    kinetic: bool = False,
    dtype: typing.Union[DTypeLike, None] = None,
) -> np.ndarray:

    if kinetic:
        max = np.max(mapping.values) + 1
    else:
        max = np.max(mapping.attrs["plasmacells"])
    if dtype is None:
        dtype = data.values.dtype

    out = np.zeros((*data.shape[:-3], max), dtype=dtype)
    mapdat = mapping.values
    assert mapdat.dtype == int
    datdat = data.values
    if out.dtype != datdat.dtype:
        if out.dtype == np.int64 and datdat.dtype == np.float64:
            datdat = (datdat + 0.5).astype(dtype)
    count = np.zeros(max, dtype=int)
    args = datdat, mapdat, out, count
    for arg in args:
        assert isinstance(arg, np.ndarray)
    out, count = to_mapped_core(*args, max)
    if out.dtype in [np.dtype(x) for x in [int, np.int32, np.int64]]:
        out //= count
    else:
        out /= count
    assert isinstance(out, np.ndarray)
    return out


def write_mapped(
    datas,
    mapping,
    fn,
    skip_first=0,
    ignore_broken=False,
    kinetic=False,
    dtype=None,
    fmt=None,
):
    """
    Write a some files using the EMC3 mapping.

    Parameters
    ----------
    datas : list of xr.DataArray
        The data to be written
    mapping : xr.DataArray
        The info about the EMC3 data mapping
    fn : str
        The filename to write to
    skip_first : None or int
        If not None, it needs to be the number of lines that are in
        the `print_before` attribute.
    ignore_broken : any
        ignored.
    kinetic : bool
        If true the data is defined also outside of the plasma region.
    dtype : any
        if not None, it needs to match the dtype of the data
    fmt : None or str
        The Format to be used for printing the data.
    """
    if skip_first is not None:
        for d in datas:
            if skip_first:
                assert "print_before" in d.attrs, (
                    'The "print_before" attribute is needed for writing. '
                    "Please ensure it is preserved or copied from the read data."
                )
                assert d.attrs["print_before"] != ""
            else:
                if "print_before" in d.attrs:
                    assert d.attrs["print_before"] == ""
    if not isinstance(datas, (list, tuple)):
        datas = [datas]
    out = [to_mapped(x, mapping, kinetic, dtype) for x in datas]
    with open(fn, "w") as f:
        for i, da in zip(out, datas):
            if "print_before" in da.attrs:
                f.write(da.attrs["print_before"])
            if fmt is None:
                tfmt = "%.4e" if dtype != int else "%d"
            else:
                tfmt = fmt
            _block_write(f, i, fmt=tfmt, bs=6)


def read_info_file(
    fn: str,
    vars: dict,
    index: str = "iteration",
    length: int = 1000,
) -> typing.List[xr.DataArray]:
    block = len(vars)
    assert block > 0
    ret: typing.List[np.ndarray] = []
    with open(fn) as f:
        while True:
            dat = _fromfile(f, dtype=float, count=block, sep=" ")
            if len(dat) == 0:
                dat = np.array(ret)[-length:]
                if len(dat) < length:
                    padded = np.empty((length, block))
                    padded[: -len(dat)] = np.nan
                    padded[-len(dat) :] = dat
                    dat = padded
                coords: typing.Mapping[typing.Hashable, typing.Any] = {
                    index: xr.DataArray(np.arange(-length + 1, 1), dims=index)
                }
                return [xr.DataArray(d, dims=index, coords=coords) for d in dat.T]
            if not len(dat) == block:
                print(dat)
                print(len(dat), block)
                print(fn)
                raise RuntimeError("Error reading file")
            ret.append(dat)


def write_info_file(fn: str, ds: xr.Dataset) -> None:
    info = files[fn.split("/")[-1]]
    fmt = info["fmt"]
    dats: typing.List[np.ndarray] = []
    for v, i in info["vars"].items():
        dats.append(ds[v].data)
        if "scale" in i:
            # Make copy, so we don't change underlying data
            dats[-1] = dats[-1] / i["scale"]
    dat = np.array(dats)
    fmtc = fmt.count("%")
    assert fmtc == len(
        dat
    ), f"Found {fmtc} format specifiers but data has {len(dat)} values. Format is {fmt}."
    # dat.shape == x, 1000
    valid_entries = np.sum(np.isfinite(dat), axis=1)
    assert len(valid_entries) == fmtc
    with open(fn, "w") as f:
        for d in dat.T[-np.max(valid_entries) :]:
            f.write(fmt % tuple(d) + "\n")


files: typing.Dict[str, typing.Dict[str, typing.Any]] = {
    "fort.70": dict(type="mapping", vars={"_plasma_map": dict()}),
    "fort.31": dict(
        type="mapped",
        skip_first=1,
        kinetic=False,
        vars={
            "ne": dict(scale=1e6, units="m$^{-3}$", long_name="Electron density"),
            "nZ%d": dict(scale=1e6, units="m$^{-3}$"),
        },
    ),
    "fort.33": dict(type="mapped", vars={"M": dict(long_name="Mach number")}),
    "fort.30": dict(
        type="mapped",
        vars={
            "Te": dict(units="eV", long_name="Electron temperature"),
            "Ti": dict(units="eV", long_name="Ion temperature"),
        },
    ),
    "CONNECTION_LENGTH": dict(
        type="mapped",
        vars={"Lc": dict(scale=1e-2, units="m", long_name="Connection length")},
    ),
    "DENSITY_A": dict(
        type="mapped",
        kinetic=True,
        vars=dict(
            nH=dict(scale=1e6, units="m$^{-3}$", long_name="Atomic deuterium density")
        ),
    ),
    "DENSITY_M": dict(
        kinetic=True,
        vars=dict(nH2=dict(scale=1e6, units="m$^{-3}$", long_name="D_2 density")),
    ),
    "DENSITY_I": dict(
        kinetic=True,
        vars={"nH2+": dict(scale=1e6, units="m$^{-3}$", long_name="D_2^+ density")},
    ),
    "TEMPERATURE_A": dict(
        kinetic=True,
        vars={"TH": dict(units="eV", long_name="Atomic hydrogen temperature")},
    ),
    "TEMPERATURE_M": dict(
        kinetic=True,
        vars={"TH": dict(units="eV", long_name="Atomic hydrogen temperature")},
    ),
    "BFIELD_STRENGTH": dict(
        type="full",
        vars={"bf_bounds": dict(units="T", long_name="Magnetic field strength")},
    ),
    "PLATES_MAG": dict(
        type="plates_mag",
        vars={"PLATES_MAG": dict(long_name="Cells that are within or behind plates")},
    ),
    # Some files - but don't know what they are
    "TEMPERATURE_I": dict(
        type="mapped",
        kinetic=True,
        vars={"TEMPERATURE_I_%d": dict()},
    ),
    "DENSITY_E_A": dict(
        type="mapped",
        kinetic=True,
        vars={"DENSITY_E_A_%d": dict()},
    ),
    "DENSITY_E_I": dict(
        type="mapped",
        kinetic=True,
        vars={"DENSITY_E_I_%d": dict()},
    ),
    "DENSITY_E_M": dict(
        type="mapped",
        kinetic=True,
        vars={"DENSITY_E_M_%d": dict()},
    ),
    "fort.1": dict(
        type="raw",
        vars={"fort.1": dict(long_name="Geometry input file")},
    ),
    "fort.2": dict(
        type="raw",
        vars={
            "fort.2": dict(
                long_name="Plasma parameters, boundary and initial conditions input file"
            )
        },
    ),
    "fort.3": dict(
        type="raw",
        vars={"fort.3": dict(long_name="Control flow input file")},
    ),
    "fort.4": dict(
        type="raw",
        vars={"fort.4": dict(long_name="Neutrals input file for EIRENE")},
    ),
    "fort.40": dict(
        type="mapped",
        vars={"fort.40_%d": dict()},
    ),
    "fort.42": dict(
        type="mapped",
        vars={"fort.42_%d": dict()},
    ),
    "fort.43": dict(
        type="mapped",
        vars={"fort.43_%d": dict()},
    ),
    "fort.46": dict(
        type="mapped",
        vars={"fort.46_%d": dict()},
    ),
    "fort.47": dict(
        type="mapped",
        vars={"fort.47_%d": dict()},
    ),
    "IMPURITY_IONIZATION_SOURCE": dict(
        type="mapped",
        vars={"IMPURITY_IONIZATION_SOURCE_%d": dict()},
    ),
    "IMPURITY_NEUTRAL": dict(
        type="mapped",
        vars={"IMPURITY_NEUTRAL_%d": dict()},
    ),
    "IMP_RADIATION": dict(
        type="mapped",
        vars={"IMP_RADIATION_%d": dict()},
    ),
    "FLUX_CONSERVATION": dict(
        type="mapped",
        vars={"FLUX_CONSERVATION_%d": dict()},
    ),
    "LG_CELL": dict(
        type="mapped",
        dtype=int,
        vars={"LG_CELL_%d": dict()},
    ),
    "STREAMING_INFO": dict(
        type="info",
        fmt="%6.2f %5.3f %10.3E %10.3E %10.3E %10.3E %10.3E",
        vars={
            "dens_change": dict(
                long_name="Relative change in density",
                scale=1e-2,
                units="",
                notes="Unlike in EMC3/pymc3 this is not percent.",
            ),
            "flow_change": dict(
                long_name="Change in Flow",
                notes="Not scaled",
            ),
            "part_balance": dict(
                long_name="Global particle balance",
                units="A",
            ),
            "dens_upstream": dict(
                long_name="Upstream Density",
                scale=1e6,
                units="m$^{-3}$",
            ),
            "dens_down_back": dict(
                long_name="Downstream Density (backward direction)",
                scale=1e6,
                units="m$^{-3}$",
            ),
            "dens_down_mean": dict(
                long_name="Downstream Density (averaged)",
                scale=1e6,
                units="m$^{-3}$",
            ),
            "dens_down_fwd": dict(
                long_name="Downstream Density (forward direction)",
                scale=1e6,
                units="m$^{-3}$",
            ),
        },
    ),
    "ENERGY_INFO": dict(
        type="info",
        fmt=("%6.1f" + " %11.4E" * 4 + "\n") * 2 + " " * 18 + 3 * " %11.4E",
        vars={
            "Te_change": dict(
                long_name="Relative change in el. temperature",
                scale=1e-2,
                units="",
                notes="Unlike in EMC3/pymc3 this is not percent.",
            ),
            "Te_upstream": dict(
                long_name="Upstream el. temperature",
                units="eV",
            ),
            "Te_down_back": dict(
                long_name="Downstream el. temperature (backward direction)", units="eV"
            ),
            "Te_down_mean": dict(
                long_name="Downstream el. temperature (averaged)", units="eV"
            ),
            "Te_down_fwd": dict(
                long_name="Downstream el. temperature (forward direction)",
                units="eV",
            ),
            "Ti_change": dict(
                long_name="Change in ion temperature",
                scale=1e-2,
                units="",
                notes="Unlike in EMC3/pymc3 this is not percent.",
            ),
            "Ti_upstream": dict(
                long_name="Upstream ion temperature",
                units="eV",
            ),
            "Ti_down_back": dict(
                long_name="Downstream ion temperature (backward direction)", units="eV"
            ),
            "Ti_down_mean": dict(
                long_name="Downstream ion temperature (averaged)", units="eV"
            ),
            "Ti_down_fwd": dict(
                long_name="Downstream ion temperature (forward direction)",
                units="eV",
            ),
            "P_loss_gas": dict(long_name="Power losses (neutral gas)", units="W"),
            "P_loss_imp": dict(long_name="Power losses (impurities)", units="W"),
            "P_loss_target": dict(long_name="Power losses (target)", units="W"),
        },
    ),
    "NEUTRAL_INFO": dict(
        type="info",
        fmt="%12.4E" + (" %11.4E" * 5),
        vars={
            "ionization_core": dict(long_name="Core ionization"),
            "ionization_edge": dict(long_name="Edge ionization"),
            "ionization_electron": dict(
                long_name="Electron energy source / ionization",
                units="eV",
            ),
            "ionization_ion": dict(
                long_name="Ion energy source / ionization",
                units="eV",
            ),
            "ionization_moment_fwd": dict(
                long_name="Forward momentum source/ ionization"
            ),
            "ionization_moment_bwk": dict(
                long_name="Backward momentum source/ ionization"
            ),
        },
    ),
    "IMPURITY_INFO": dict(
        type="info",
        fmt="%12.4E %11.4E",
        vars={
            "TOTAL_FLX": dict(long_name="Total impurity flux"),
            "TOTAL_RAD": dict(long_name="Total radiation", units="W"),
        },
    ),
    "ADD_SF_N0": dict(
        type="surfaces",
        vars={
            plate_prefix + "phi": dict(units="radian"),
            plate_prefix + "R": dict(units="m"),
            plate_prefix + "z": dict(units="m"),
        },
    ),
    "GRID_3D_DATA": dict(
        type="geom",
        vars={
            "R_bounds": dict(units="m"),
            "z_bounds": dict(units="m"),
            "phi_bounds": dict(units="radian"),
        },
    ),
    "PARTICLE_DEPO": dict(
        type="depo",
        vars={
            "surftype_ne": dict(description="True means +1, False means -1"),
            "flux_ne": dict(long_name="Outflux of particles", units="s^-1"),
            "PARTICLE_DEPO_%d": dict(),
        },
    ),
    "ENERGY_DEPO": dict(
        type="depo",
        vars={
            "surftype_Te": dict(description="True means +1, False means -1"),
            "flux_P": dict(long_name="Outflux of energy", units="W"),
            "ENERGY_DEPO_%d": dict(),
        },
    ),
    "TARGET_PROFILES": dict(
        type="target_flux",
        vars={
            "f_n": dict(long_name="Particle flux"),
            "f_E": dict(units="W/m²", scale=1e4, long_name="Energy flux"),
            "avg_n": dict(units="m^-3", scale=1e6, long_name="Averge density"),
            "avg_Te": dict(units="eV", long_name="Average electron temperature"),
            "avg_Ti": dict(units="eV", long_name="Average ion temperature"),
        },
    ),
}

if False:
    _files_bak = files.copy()
    for k in files:
        _files_bak[k] = files[k].copy()
        if isinstance(files[k], dict):
            for l in files[k]:
                try:
                    _files_bak[k][l] = files[k][l].copy()
                except:  # noqa: E722
                    try:
                        _files_bak[k][l] = files[k][l][:]
                    except:  # noqa: E722
                        pass
else:
    _files_bak = files


def read_fort_file_pub(
    fn: str, ds: typing.Union[None, xr.Dataset, xr.DataArray] = None, **opts
) -> xr.Dataset:
    """
    Read a EMC3 simulation file. The expected content is derived from
    the filename.

    Parameters
    ----------
    fn : str
        The location of the file to read
    ds : xr.DataArray or xr.Dataset or None
        The mapping or a dataset containing a mapping or None. If one
        is needed and none is provided, it is read from the folder.
    opts : dict
        Additional options depending on the type of file that is read.

    Returns
    -------
    xr.Dataset
        The read variable is set. If a Dataset was given, the newly
        read entries from the file are added. Otherwise a new Dataset
        is returned with the read variables added.
    """
    filename = fn.split("/")[-1]
    defaults = files[filename].copy()
    defaults.update(opts)
    type = defaults.get("type", "mapped")
    if type == "info":
        if not isinstance(ds, xr.Dataset):
            ds = xr.Dataset()
        return read_fort_file(ds, fn, **defaults)
    ds = ensure_mapping("/".join(fn.split("/")[:-1]), ds, type == "mapped")
    assert isinstance(ds, xr.Dataset)
    return read_fort_file(ds, fn, **defaults)


def read_fort_file(ds: xr.Dataset, fn: str, type: str = "mapped", **opts) -> xr.Dataset:
    """
    Read an EMC3 simulation file and add to a dataset.

    """
    assert files == _files_bak
    datas = None
    vars = opts.pop("vars")
    if type == "mapping":
        ds["_plasma_map"] = read_mappings(fn, ds.R_bounds.data.shape[:3])
    elif type == "mapped":
        ds = ensure_mapping(_dir_of(fn), ds)
        datas = read_mapped(fn, ds["_plasma_map"], **opts, squeeze=False)
        opts = {}
    elif type == "full":
        datas = [read_magnetic_field(fn, ds)]
    elif type == "plates_mag":
        datas = [read_plates_mag(fn, ds)]
    elif type == "geom":
        ds_ = read_locations(_dir_of(fn))
        ds = ds.assign_coords(ds_.coords)
        assert opts == {}, "Unexpected arguments: " + ", ".join(
            [f"{k}={v}" for k, v in opts.items()]
        )
    elif type == "info":
        opts.pop("fmt")
        if "iteration" in ds.dims and "length" not in opts:
            opts["length"] = len(ds["iteration"])
        datas = read_info_file(fn, vars, **opts)
        opts = {}
    elif type == "surfaces":
        ds_ = read_add_sf_n0(fn)
        ds = ds.assign_coords(ds_.coords)
        for k in ds_:
            ds[k] = ds_[k]
        datas = None
    elif type == "target_flux":
        ds_ = get_plates(fn, False)
        ds = ds.assign_coords(ds_.coords)
        for k in ds_:
            ds[k] = ds_[k]
        datas = None
    elif type == "raw":
        datas = [read_raw(fn)]
    elif type == "depo":
        datas = read_depo_raw(ds, fn)
    else:
        raise RuntimeError(f"Unexpected type {type}")
    assert files == _files_bak
    assert opts == {}, "Unexpected arguments: " + ", ".join(
        [f"{k}={v}" for k, v in opts.items()]
    )
    if datas is None:
        return ds
    vars = vars.copy()
    assert files == _files_bak
    assert opts == {}
    keys = [k for k in vars]
    if "%" in keys[-1]:
        key = keys[-1]
        flexi = vars.pop(key)
        for i in range(len(vars), len(datas)):
            vars[key % i] = flexi

    assert files == _files_bak
    assert len(vars) == len(
        datas
    ), f"in file {fn} we found {len(datas)} fields but only {len(vars)} vars are given!"
    assert files == _files_bak
    for (var, varopts), data in zip(vars.items(), datas):
        ds[var] = data
        varopts = varopts.copy()
        scale = varopts.pop("scale", 1)
        if scale != 1:
            ds[var].data *= scale

        attrs = varopts.pop("attrs", {})
        attrs["xemc3_type"] = type
        for k in "long_name", "units", "notes", "description":
            if k in varopts:
                attrs[k] = varopts.pop(k)

        ds[var].attrs.update(attrs)
        assert (
            varopts == {}
        ), f"variable {var} has options {varopts} but didn't expect anything"
    assert files == _files_bak
    return ds


def guess_type(ds: xr.Dataset, key: typing.Hashable) -> str:
    data = ds[key]
    assert isinstance(key, str)
    try:
        ret = data.attrs["xemc3_type"]
        assert isinstance(ret, str)
        return ret
    except KeyError:
        pass
    if key == "_plasma_map":
        return "mapping"
    nameparts = key.split("_")
    if len(nameparts) == 2:
        if nameparts[0] in ["R", "z", "phi"] and nameparts[1] in [
            "bounds",
            "corners",
        ]:
            return "geom"
    fulldims = len(
        {"R", "phi", "z", "delta_R", "delta_z", "delta_phi"}.intersection(ds[key].dims)
    )
    if fulldims == 6:
        return "full"
    phidims = len({"phi", "delta_phi"}.intersection(ds[key].dims))
    if phidims == fulldims == 2:
        return "geom"
    if data.dtype == bool:
        return "plates_mag"
    if data.dtype == int:
        return "mapping"
    return "mapped"


def guess_kinetic(ds: xr.Dataset, key: typing.Hashable) -> bool:
    pm = ds._plasma_map
    assert ds[key].dtype in [float, np.float32, np.float16]
    return bool(np.any(np.isfinite(ds[key].data[(..., pm == pm.attrs["plasmacells"])])))


def archive(ds: xr.Dataset, fn: str, geom: bool = False, mapping: bool = True) -> None:
    arch = xr.Dataset()
    arch.attrs = ds.attrs
    for k in list(ds.coords) + list(ds):
        type = guess_type(ds, k)
        if type == "mapped":
            kinetic = guess_kinetic(ds, k)
            arch[k] = (
                *ds[k].dims[:-3],
                "kinetic_map" if kinetic else "plasma_map",
            ), to_mapped(ds[k], ds._plasma_map, guess_kinetic(ds, k))
        elif type in ("geom", "full"):
            if not geom:
                continue
            arch[k] = (
                tuple(
                    [
                        f"{x}_plus1"
                        for x in ds[k].dims
                        if x[:6] != "delta_"  # type: ignore
                    ]
                ),
                from_interval(ds[k]),
            )
        elif type == "plates_mag":
            if not geom:
                continue
            arch[k] = ds[k]
        elif type == "mapping":
            if not mapping:
                continue
            arch[k] = ds[k]
        else:
            arch[k] = ds[k]
        arch[k].attrs = ds[k].attrs
    arch.to_netcdf(
        fn,
        encoding={
            i: {"zlib": True, "complevel": 9} for i in list(arch) + list(arch.coords)
        },
    )
    print(f"done with {fn}")
    pass


def load_all(path: str, ignore_missing: bool = None) -> xr.Dataset:
    """
    Load all data from a path and return as dataset

    Parameters
    ----------
    path : str
         Directory that contains files to be read
    ignore_missing : None or bool
         True: ignore missing files.
         False: raise exceptions if a file is not found.
         None: use default option for that file.

    Returns
    -------
    xr.Dataset
        A dataset with all the simulation data. Unless
        ``ignore_missing=False`` has been set, some data might be
        missing because it was not found. However
        ``ignore_missing=False`` has the disadvantage that files added
        in later versions will cause errors if the files are not
        present.
    """
    ds = read_locations(path)
    for fn, opts in files.items():
        opts = opts.copy()
        try:
            ds = read_fort_file(ds, f"{path}/{fn}", **opts)
        except FileNotFoundError:
            if ignore_missing is None:
                if not opts.get("ignore_missing", True):
                    raise
            elif not ignore_missing:
                raise
    return ds


def load_any(path: str, *args, **kwargs) -> xr.Dataset:
    """
    Read a file or directory. For possible kwargs see the respective
    functions that are called:

    For a directory xemc3.load.all is called.

    xemc3.load.plates for the "TARGET_PROFILES" file.

    xemc3.load.file otherwise.
    """
    if os.path.isdir(path):
        return load_all(path, *args, **kwargs)
    return read_fort_file_pub(path, *args, **kwargs)


def write_fort_file(ds, dir, fn, type="mapped", **opts):
    if type == "mapping":
        assert list(opts["vars"].keys()) == ["_plasma_map"]
        write_mappings(ds["_plasma_map"], f"{dir}/{fn}")
    elif type == "mapped":
        write_mapped_nice(ds, dir, fn)
    elif type == "full":
        vars = opts.pop("vars")
        assert len(vars) == 1
        for var in vars:
            assert var == "bf_bounds"
            write_magnetic_field(f"{dir}/{fn}", ds)
    elif type == "plates_mag":
        vars = opts.pop("vars")
        assert list(vars.keys()) == ["PLATES_MAG"]
        write_plates_mag(f"{dir}/{fn}", ds)
    elif type == "info":
        write_info_file(f"{dir}/{fn}", ds)
    elif type == "geom":
        assert fn == "GRID_3D_DATA"
        write_locations(ds, f"{dir}/{fn}")
    elif type == "raw":
        vars = opts.pop("vars")
        assert len(vars) == 1
        write_raw(ds[list(vars.keys())[0]], f"{dir}/{fn}")
    elif type == "depo":
        vars = get_vars_for_file(ds, opts.pop("vars"))
        datas = [ds[k] for k, _ in vars]
        write_depo_raw(datas, f"{dir}/{fn}")
    elif type in ["surfaces", "target_flux"]:
        print(f"writing {fn} is not yet implemented. (PRs welcome)")
    else:
        raise RuntimeError(f"Unexpected type {type}")


def write_all_fortran(ds, dir):
    """
    Write all files to directory dir

    Parameters
    ----------
    ds : xr.Dataset
        The xemc3 dataset
    dir : str
        The directory to write the files to
    """
    write_locations(ds, f"{dir}/GRID_3D_DATA")
    for fn, opts in files.items():
        try:
            get_vars_for_file(ds, fn)
        except KeyError:
            pass
        else:
            write_fort_file(ds, dir, fn, **opts)

    # if False:
    #     ds["phi"] = read_mapped(path + "/POTENTIAL", map, skip_first=0, kinetic=False)[
    #         0
    #     ]
    #     ds["phi"].attrs["units"] = "V"
    #     ds["phi"].attrs["long_name"] = "Potential"

    #     ds["current"] = read_mapped(
    #         path + "/CURRENT_V", map, skip_first=0, kinetic=False
    #     )[0]
    #     ds["current"].attrs["units"] = "A"
    #     ds["current"].attrs["long_name"] = "parallel current"

    #     ds["res"] = (
    #         read_mapped(path + "/RESISTIVITY", map, skip_first=0, kinetic=False)[0]
    #         * 1e-4
    #     )
    #     ds["res"].attrs["units"] = "$\Omega$ m$^2$"
    #     ds["res"].attrs["long_name"] = "$\int \sigma_{||}^{-1}dl$"

    # # SMalpha                        fort.46                        0 6.2415093237819604e+22 $S_{M,transp.}$ [kg m$^{-2}$ s$^{-2}$]
    #     ds["SMalpha"]=read_mapped(path + "/fort.46", map, skip_first=0, kinetic=False)[0]*6.2415093237819604e+22
    #     ds["SMalpha"].attrs["units"] = "$S_{M,transp.}$ [kg m$^{-2}$ s$^{-2}$]"
    #     ds["SMalpha"].attrs["long_name"] = "$S_{M,transp.}$ [kg m$^{-2}$ s$^{-2}$]"

    # # SM                             fort.47                        0 6.2415093237819604e+22 $S_{M,CX}$ [kg m$^{-2}$ s$^{-2}$]
    #     ds["SM"]=read_mapped(path + "/fort.47", map, skip_first=0, kinetic=False)[0]*6.2415093237819604e+22
    #     ds["SM"].attrs["units"] = "$S_{M,CX}$ [kg m$^{-2}$ s$^{-2}$]"
    #     ds["SM"].attrs["long_name"] = "$S_{M,CX}$ [kg m$^{-2}$ s$^{-2}$]"

    # # imprad                         IMP_RADIATION                  0 -1000000.0 $P_{rad,imp}$ [W m$^{-3}$]
    #     ds["imprad"]=read_mapped(path + "/IMP_RADIATION", map, skip_first=0, kinetic=False)[0]*-1000000.0
    #     ds["imprad"].attrs["units"] = "$P_{rad,imp}$ [W m$^{-3}$]"
    #     ds["imprad"].attrs["long_name"] = "$P_{rad,imp}$ [W m$^{-3}$]"

    # # Simp                           IMPURITY_IONIZATION_SOURCE     0 1.0 S$_Z$ []
    #     ds["Simp"]=read_mapped(path + "/IMPURITY_IONIZATION_SOURCE", map, skip_first=0, kinetic=False)[0]
    #     ds["Simp"].attrs["units"] = "S$_Z$ []"
    #     ds["Simp"].attrs["long_name"] = "S$_Z$ []"

    # # flux_conservation              FLUX_CONSERVATION              0 1.0 flux []
    #     ds["flux_conservation"]=read_mapped(path + "/FLUX_CONSERVATION", map, skip_first=0, kinetic=False)[0]
    #     ds["flux_conservation"].attrs["units"] = "flux []"
    #     ds["flux_conservation"].attrs["long_name"] = "flux []"

    # # R                              RECOMBINATION                  0 1.6021765699999998e-13 R [s$^{-1}$ m$^{-3}$]
    #     ds["R"]=read_mapped(path + "/RECOMBINATION", map, skip_first=0, kinetic=False)[0]*1.6021765699999998e-13
    #     ds["R"].attrs["units"] = "R [s$^{-1}$ m$^{-3}$]"
    #     ds["R"].attrs["long_name"] = "R [s$^{-1}$ m$^{-3}$]"

    return ds



"""
A python library for reading EMC3 simulation as xarray data.

Additionally also provides some basic routines for plotting and
analysing the data. The data is mostly in SI units, with the exception
of temperatures that are in eV.
"""

__all__ = ["load", "utils", "write"]

from . import load, write
from .core.dataset import EMC3DatasetAccessor


try:
    from importlib.metadata import (  # type: ignore
        PackageNotFoundError as _PackageNotFoundError,
        version as _version,
    )
except ModuleNotFoundError:
    from importlib_metadata import (  # type: ignore
        PackageNotFoundError as _PackageNotFoundError,
        version as _version,
    )
try:
    __version__ = _version(__name__)
except _PackageNotFoundError:
    from setuptools_scm import get_version as _get_version  # type: ignore

    __version__ = _get_version(root="..", relative_to=__file__)



import warnings
from timeit import timeit

import numpy as np
import pytest
import xarray as xr

from ..core import dataset
from . import gen_ds

try:
    import matplotlib  # type: ignore
except ImportError:
    matplotlib = None


def test_sel():
    ds = xr.Dataset()

    ds.emc3["r_corners"] = "r", np.linspace(0, 4, 5)
    ds.emc3["x"] = "r", np.linspace(0, 3, 4)

    for i in range(5):
        dss = ds.emc3.sel(r=i)
        assert dss["r_bounds"] == i
        assert dss["x"] == min(i, 3)
    for i in np.random.random(10) * 4:
        dss = ds.emc3.sel(r=i)
        assert dss["r_bounds"] == i
        assert dss["x"] == int(i)


def test_isel():
    ds = xr.Dataset()
    ds.emc3["r_corners"] = "r", np.linspace(0, 4, 5)
    ds.emc3["x"] = "r", np.linspace(0, 3, 4)

    for i in range(5):
        dss = ds.emc3.isel(r=i)
        assert dss["r_bounds"] == i
        assert dss["x"] == min(i, 3)
    for i in np.random.random(10) * 4:
        dss = ds.emc3.isel(r=i)
        assert dss["r_bounds"] == i
        assert dss["x"] == int(i)


def test_isel2():
    ds = xr.Dataset()
    rf = 2
    xf = 3
    ds.emc3["r_corners"] = "r", np.linspace(0, 4 * rf, 5)
    ds.emc3["x"] = "r", np.linspace(0, 3 * xf, 4)

    for i in range(5):
        dss = ds.emc3.isel(r=i)
        assert dss["r_bounds"] == i * rf
        assert dss["x"] == min(i, 3) * xf
    for i in np.random.random(10) * 4:
        dss = ds.emc3.isel(r=i)
        assert dss["r_bounds"] == i * rf
        assert dss["x"] == int(i) * xf


def test_isel_multi():
    ds = xr.Dataset()
    rf = 2
    zf = 5
    xf = 3
    rr = np.linspace(0, 4 * rf, 5)
    zr = np.linspace(0, 7 * zf, 8)
    ds.emc3["r_corners"] = ("r", "z"), (rr[:, None] + zr[None, :])
    ds.emc3["x"] = "r", np.linspace(0, 3 * xf, 4)
    ds.emc3["y"] = "z", np.linspace(0, 6 * xf, 7)
    # ds.emc3["k"] = ("r","z"), np.linspace(0, 6 * xf, 7)

    for i in range(5):
        dss = ds.emc3.isel(r=i)
        assert all(dss.emc3["r_corners"] == i * rf + zr)
        assert dss["x"] == min(i, 3) * xf
        assert all(dss["y"] == ds["y"])
    for j in range(8):
        dss = ds.emc3.isel(z=j)
        assert all(dss.emc3["r_corners"] == j * zf + rr)
    for i in np.random.random(10) * 4:
        dss = ds.emc3.isel(r=i)
        assert np.allclose(dss.emc3["r_corners"], i * rf + zr)
        assert dss["x"] == int(i) * xf


def test_isel_multi2():
    ds = xr.Dataset()
    rf = 2
    zf = 5
    xf = 3
    rr = np.linspace(0, 4 * rf, 5)
    zr = np.linspace(0, 7 * zf, 8)
    ds_ = ds.copy()
    ds_.emc3["r_corners"] = ("r", "z"), (rr[:, None] + zr[None, :])
    ds = ds.assign_coords({"r_bounds": ds_["r_bounds"]})
    ds.emc3["x"] = "r", np.linspace(0, 3 * xf, 4)
    ds.emc3["y"] = "z", np.linspace(0, 6 * xf, 7)
    # ds.emc3["k"] = ("r","z"), np.linspace(0, 6 * xf, 7)

    for i in range(5):
        dss = ds.emc3.isel(r=i)
        assert all(dss.emc3["r_corners"] == i * rf + zr)
        assert dss["x"] == min(i, 3) * xf
        assert all(dss["y"] == ds["y"])
    for j in range(8):
        dss = ds.emc3.isel(z=j)
        assert all(dss.emc3["r_corners"] == j * zf + rr)
    for i in np.random.random(10) * 4:
        dss = ds.emc3.isel(r=i)
        assert np.allclose(dss.emc3["r_corners"], i * rf + zr)
        assert dss["x"] == int(i) * xf


def test_sel_multi2():
    ds = xr.Dataset()
    rf = 2
    zf = 5
    xf = 3
    rr = np.linspace(0, 4 * rf, 5)
    zr = np.linspace(0, 7 * zf, 8)
    ds_ = ds.copy()
    ds_.emc3["r_corners"] = ("r", "z"), (rr[:, None] + zr[None, :])
    ds = ds.assign_coords({"r_bounds": ds_["r_bounds"]})
    ds.emc3["x"] = "r", np.linspace(0, 3 * xf, 4)
    ds.emc3["z_corners"] = "z", zr
    # ds.emc3["k"] = ("r","z"), np.linspace(0, 6 * xf, 7)

    for j in range(8):
        dss = ds.emc3.sel(z=j * zf)
        assert all(dss.emc3["r_corners"] == j * zf + rr)
    for i in np.random.random(10) * 7:
        dss = ds.emc3.sel(z=i * zf)
        assert np.allclose(dss.emc3["r_corners"], i * zf + rr)


def test_mean_dtype():
    ds = xr.Dataset()
    ds["pos"] = [1, 2, 3]
    ds["data"] = ("pos", "time"), [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]
    ds["var"] = "pos", [2, 3, 4]
    ds2 = ds.emc3.mean_time()
    assert all(ds2["var"] == ds["var"])
    assert ds2["var"].dtype == ds["var"].dtype


class Test_eval_at_rpz(object):
    def setup(self, shape=None, **kwargs):
        if not isinstance(shape, tuple):
            shape = None
        self.shape = shape or (2, 20, 4)
        self.geom = gen_ds.rotating_circle(5, **kwargs)
        self.ds = self.geom.gends(self.shape)
        self.dims = self.ds["_plasma_map"].dims
        self.dphi = 2 * np.pi / self.shape[2] / 5

    def rand_rpt(self, i):
        slc = [slice(None)]
        if isinstance(i, np.ndarray):
            slc += [None] * len(i.shape)
        else:
            slc += [None]
        tmp = (
            np.random.random((3, i))
            * np.array([self.geom.r * 0.9, np.pi * 2, np.pi * 2])[tuple(slc)]
        )
        r, p, t = tmp
        return r, p, t

    def rand(self, i):
        return self.geom.rpt_to_rpz(*self.rand_rpt(i))

    def phi_test_value(self, a, b):
        dphi = self.dphi
        phi = np.linspace(dphi / 2, 2 * np.pi / 5 - dphi / 2, self.shape[2])
        self.ds["var"] = self.dims, np.zeros(self.shape) + phi
        for r, p, z in [self.rand(a) for _ in range(b)]:
            exp = (np.round((p - dphi / 2) / dphi) % self.shape[2]) * dphi + dphi / 2
            got = self.ds.emc3.evaluate_at_rpz(r, p, z, "var", updownsym=self.geom.sym)[
                "var"
            ]
            assert np.allclose(
                exp,
                got,
            )

    def phi_test_index(self, a, b):
        dphi = self.dphi
        for r, p, z in [self.rand(a) for _ in range(b)]:
            assert np.allclose(
                np.round((p - dphi / 2) / dphi) % self.shape[2],
                self.ds.emc3.evaluate_at_rpz(r, p, z, updownsym=self.geom.sym)["phi"],
            )

    def test_phi_single1(self):
        self.setup()
        self.phi_test_value(1, 20)

    def test_phi_single2(self):
        self.setup()
        self.phi_test_index(1, 20)

    def test_phi_some1(self):
        self.setup()
        self.phi_test_value(2, 5)

    def test_phi_some2(self):
        self.setup()
        self.phi_test_value(5, 1)

    def test_phi_some3(self):
        self.setup()
        self.phi_test_index(4, 1)

    def r_test_value(self, a, b):
        dr = self.geom.r / self.shape[0]
        r = np.linspace(dr / 2, self.geom.r - dr / 2, self.shape[0])
        self.ds["var"] = self.dims, np.zeros(self.shape) + r[:, None, None]
        for r, p, t in [self.rand_rpt(a) for _ in range(b)]:
            expl = (np.round((r - dr * 0.51) / dr)) * dr + dr * 0.49
            exp = (
                np.round((r / np.cos(np.pi / self.shape[1]) - dr * 0.48) / dr)
            ) * dr + dr * 0.52
            R, p, z = self.geom.rpt_to_rpz(r, p, t)
            got = self.ds.emc3.evaluate_at_rpz(R, p, z, "var", updownsym=self.geom.sym)[
                "var"
            ].data
            if a == 1:
                isgood = expl < got < exp
            else:
                isgood = all([a < b < c for a, b, c in zip(expl, got, exp)])
            if not isgood:
                for i, t in enumerate([a < b < c for a, b, c in zip(expl, got, exp)]):
                    if not t:
                        import matplotlib.pyplot as plt  # type: ignore

                        self.ds.emc3.plot_rz(key=None, phi=p[i] % (np.pi * 2 / 5))
                        plt.plot(R[i], z[i], "rx")
                        plt.show()
            assert isgood, f"""{expl} <  {got} < {exp}
{R}, {p}, {z}
{r}, {p}, {t})"""

    def theta_test_value(self, a, b):
        dt = 2 * np.pi / self.shape[1]
        t = np.linspace(0, 2 * np.pi - dt, self.shape[1])
        self.ds["var"] = self.dims, np.zeros(self.shape) + t[None, :, None]
        for r, p, t in [self.rand_rpt(a) for _ in range(b)]:
            # No test within phi, as then we need to calculate where we end up, which is non-trivial
            p = np.zeros_like(p)
            exp = (np.round(((t)) / dt) % self.shape[1]) * dt
            R, p, z = self.geom.rpt_to_rpz(r, p, t)
            got = self.ds.emc3.evaluate_at_rpz(R, p, z, "var", updownsym=self.geom.sym)[
                "var"
            ].data
            isgood = np.allclose(got, exp)
            if not isgood:
                print(got / dt, exp / dt)
                print(R, p, z)
                print(r, p, t / dt)
                assert False

    def test_r_single(self):
        self.setup()
        self.r_test_value(1, 20)

    def test_r_some(self):
        self.setup()
        self.r_test_value(2, 3)
        self.r_test_value(3, 4)

    def test_theta_single(self):
        self.setup()
        self.theta_test_value(1, 20)

    def test_trace_line(self):
        self.setup(shape=(5, 30, 10))
        var = np.zeros(self.shape)
        for i in range(5):
            var[2, i * self.shape[1] // 5, :] = 1

        p = np.linspace(0, 2 * np.pi, 100)
        t = np.zeros_like(p)
        r = self.geom.r * 0.5 + np.zeros_like(p)
        R, p, z = self.geom.rpt_to_rpz(r, p, t)

        self.ds["var"] = self.dims, var

        # plt = self.ds.emc3.plot(
        #     "var", updownsym=self.geom.sym, periodicity=1  # self.geom.period
        # )
        # import mayavi.mlab as mlab  # type: ignore

        # for i in range(5):
        #     mlab.plot3d(
        #         *self.geom.rpz_to_xyz(
        #             *self.geom.rpt_to_rpz(r, p, t + i * np.pi * 2 / 5)
        #         ),
        #         opacity=0.9,
        #     )
        # plt.show()
        got = self.ds.emc3.evaluate_at_rpz(R, p, z, "var", updownsym=self.geom.sym)[
            "var"
        ].data
        assert np.allclose(got, 1), f"got {got} but expected 1"

    def test_trace_line_updown(self):
        self.setup(shape=(5, 60, 10), sym=True)
        var = np.zeros(self.shape)
        for i in range(5):
            var[2, i * self.shape[1] // 5, :] = 1

        p = np.linspace(0, 2 * np.pi, 100)
        t = np.zeros_like(p)  # + np.pi / self.shape[1]
        r = self.geom.r * 0.5 + np.zeros_like(p)
        R, p, z = self.geom.rpt_to_rpz(r, p, t)
        self.ds["var"] = self.dims, var

        # plt = self.ds.emc3.plot(
        #     "var", updownsym=self.geom.sym, periodicity=1  # self.geom.period
        # )
        # import mayavi.mlab as mlab  # type: ignore

        # for i in range(5):
        #     mlab.plot3d(
        #         *self.geom.rpz_to_xyz(
        #             *self.geom.rpt_to_rpz(r, p, t + i * np.pi * 2 / 5)
        #         ),
        #         opacity=0.9,
        #     )
        # plt.show()
        got = self.ds.emc3.evaluate_at_rpz(R, p, z, "var", updownsym=self.geom.sym)[
            "var"
        ].data
        assert np.allclose(got, 1), f"got {got} but expected 1"

    def test_cached_eval(self):
        a = 200
        b = 1
        self.setup()  # (10, 20, 30))
        dphi = self.dphi
        for r, p, z in [self.rand(a) for _ in range(b)]:
            exp = np.round((p - dphi / 2) / dphi) % self.shape[2]
            got = self.ds.emc3.evaluate_at_rpz(
                r, p, z, updownsym=self.geom.sym, delta_phi=dphi
            )["phi"]
            assert np.allclose(
                exp, got
            ), f"Expected \n{exp} but got \n{got.data} \n{p/dphi} % {self.shape[2]}"
            assert got.dims == ("dim_0",)

    def test_cached_eval_perf(self):
        self.setup((2, 20, 30))
        dphi = self.dphi
        ds = self.ds
        r, p, z = self.rand(40)
        cached = timeit(
            """ds.emc3.evaluate_at_rpz(
                r, p, z, updownsym=self.geom.sym, delta_phi=dphi
        )""",
            number=1,
            globals=locals(),
        )

        slow = timeit(
            """ds.emc3.evaluate_at_rpz(
                r, p, z, updownsym=self.geom.sym
        )""",
            number=1,
            globals=locals(),
        )
        print(slow, cached, slow / cached)
        assert (
            slow > cached
        ), f"Expected the cached version to be faster then the non-cached {slow} vs {cached}. Note that this might sometimes fail. Increase the number of samples to avoid that."

    def test_nan_value(self):
        self.setup()
        dt = 2 * np.pi / self.shape[1]
        t = np.linspace(0, 2 * np.pi - dt, self.shape[1])
        self.ds["var"] = self.dims, np.zeros(self.shape) + t[None, :, None]
        r, p, t = self.rand_rpt(6)
        # No test within phi, as then we need to calculate where we end up, which is non-trivial
        p = np.zeros_like(p)
        exp = (np.round(((t)) / dt) % self.shape[1]) * dt
        dat = np.array([r, p, t, exp])
        dat[:, [0, 2, 5]] = np.nan
        r, p, t, exp = dat
        R, p, z = self.geom.rpt_to_rpz(r, p, t)
        got = self.ds.emc3.evaluate_at_rpz(R, p, z, "var", updownsym=self.geom.sym)[
            "var"
        ].data
        isgood = all(np.isnan(exp) == np.isnan(got))
        assert isgood, f"""
        {exp / dt} {got / dt}
        {R} {p} {z}
        {r} {p} {t / dt}
"""

    def test_at_xyz(self):
        self.setup()
        dt = 2 * np.pi / self.shape[1]
        t = np.linspace(0, 2 * np.pi - dt, self.shape[1])
        self.ds["var"] = self.dims, np.zeros(self.shape) + t[None, :, None]
        r, p, t = self.rand_rpt(6)
        # No test within phi, as then we need to calculate where we end up, which is non-trivial
        p = np.zeros_like(p)
        exp = (np.round(((t)) / dt) % self.shape[1]) * dt
        Rpz = self.geom.rpt_to_rpz(r, p, t)
        xyz = self.geom.rpz_to_xyz(*Rpz)
        got = self.ds.emc3.evaluate_at_xyz(*xyz, "var", updownsym=self.geom.sym)[
            "var"
        ].data
        assert all(np.isclose(got, exp))

    def test_at_dataarrays(self):
        self.setup()
        t = np.linspace(0, 2 * np.pi, self.shape[2])
        self.ds["var"] = self.dims, np.zeros(self.shape) + t
        ds = xr.Dataset(coords=dict(x=np.linspace(0, 3, 5), y=np.linspace(0, 2, 6)))
        x = ds.x
        y = ds.y
        for z in 0, x * y:
            result = self.ds.emc3.evaluate_at_xyz(
                x, y, z, "var", updownsym=self.geom.sym
            )
            assert "x" in result.dims
            assert "y" in result.dims
            assert all(result.coords["x"] == x)
            assert all(result.coords["y"] == y)
            assert result["var"].dims == ("x", "y")
            assert result["var"].shape == (5, 6)

    def test_plot_rz(self):
        """
        Ensure no error is raised.
        """
        if matplotlib is None:
            pytest.skip("matplotlib missing")
        self.setup()
        ret = self.ds.emc3.plot_rz(key=None, phi=0.3)
        assert isinstance(ret, matplotlib.collections.QuadMesh)

    def test_plot_Rz(self):
        """
        Ensure no error is raised.
        """
        if matplotlib is None:
            pytest.skip("matplotlib missing")
        self.setup()
        with warnings.catch_warnings(record=True) as w:
            ret = self.ds.emc3.plot_Rz(key=None, phi=0.3)
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "plot_rz" in str(w[-1].message)
        assert isinstance(ret, matplotlib.collections.QuadMesh)


if __name__ == "__main__":
    Test_eval_at_rpz().test_nan_value()



import tempfile

import numpy as np
from hypothesis import given, settings

import xemc3

from . import gen_ds


def assert_ds_are_equal(d1, d2, check_attrs=True, rtol=1e-2, atol=1e-6):
    """
    d1 : old dataset
    d2 : new dataset
    """
    d1k = [k for k in d1] + [k for k in d1.coords]
    d2k = [k for k in d2] + [k for k in d2.coords]
    if not set(d1k) == set(d2k):
        raise AssertionError(f"{d1.keys()} != {d2.keys()}")
    for k in d1k:
        assert_da_are_equal(
            d1[k], d2[k], k, check_attrs, rtol, 1e-3 if k.endswith("_change") else atol
        )


def assert_da_are_equal(d1, d2, k, check_attrs, rtol, atol):
    if d1.dtype == float:
        slc = np.isfinite(d1.data)
    else:
        slc = slice(None)
    if (
        d1.dtype == float
        and (
            d1.shape != d2.shape
            or not np.isclose(d1.data[slc], d2.data[slc], rtol=rtol, atol=atol).all()
        )
    ) or (d1.dtype != float and (d1.shape != d2.shape or np.any(d1.data != d2.data))):
        raise AssertionError(
            f"""var {k} is changed.

Before: {d1.shape}: {d1.data.flatten()}

After: {d2.shape}: {d2.data.flatten()}

np.isclose: {np.isclose(d1, d2 ,rtol=rtol).flatten()}"""
        )
    if check_attrs:
        d1a = d1.attrs.copy()
        key = "xemc3_type"
        if key not in d1a and key in d2.attrs:
            d1a[key] = d2.attrs[key]

        assert d1a == d2.attrs, f"attributes changed for {k}: {d1a} != {d2.attrs}"


setting = gen_ds.setting


@settings(**setting)  # type: ignore
@given(gen_ds.hypo_shape())
def test_write_load_simple(shape):
    ds = gen_ds.gen_ds(shape)
    with tempfile.TemporaryDirectory() as dir:
        # print(ds)
        # print(ds["_plasma_map"])
        xemc3.write.fortran(ds, dir)
        dl = xemc3.load(dir)
        assert_ds_are_equal(ds, dl, False, 1e-2, 1e-2)
    with tempfile.TemporaryDirectory() as dir:
        xemc3.write.fortran.all(dl, dir)
        dn = xemc3.load.all(dir)
        da = xemc3.load.mapped_raw(dir + "/DENSITY_A", dn, kinetic=True)
        assert not isinstance(da, list), "test with squeeze failed"
        assert np.allclose(da, dn["nH"] / 1e6), "reading mapped_raw gives wrong data"
        da = xemc3.load.mapped_raw(dir + "/DENSITY_A", dn, kinetic=True, squeeze=False)
        assert isinstance(da, list), "squeeze=False gives list"
        assert np.allclose(da[0], dn["nH"] / 1e6), "mapped_raw gives wrong data"
        for var, fn in ("nH", "DENSITY_A"), ("bf_bounds", "BFIELD_STRENGTH"):
            for withmaps in True, False:
                da = xemc3.load.var(dir, var, dn["_plasma_map"] if withmaps else None)
                assert np.allclose(da[var], dn[var]), (
                    "load.var fails with" + ("" if withmaps else "out") + " map fails"
                )
                da = xemc3.load.file(
                    f"{dir}/{fn}", dn["_plasma_map"] if withmaps else None
                )
                assert np.allclose(da[var], dn[var]), (
                    "load.file fails with" + ("" if withmaps else "out") + " map fails"
                )

        # print(xemc3.core.load.files)
        assert_ds_are_equal(dl, dn, True, 1e-4)


@settings(**setting)  # type: ignore
@given(gen_ds.hypo_shape(200))
def test_write_load_full(shape):
    ds = gen_ds.gen_full(shape)
    # if True:
    #    dir = "xemc3/test/test"
    with tempfile.TemporaryDirectory() as dir:
        xemc3.write.fortran(ds, dir)
        dl = xemc3.load(dir)
        assert_ds_are_equal(ds, dl, True, 1e-2, 1e-2)
    with tempfile.TemporaryDirectory() as dir:
        xemc3.write.fortran.all(dl, dir)
        dn = xemc3.load.all(dir)
        # print(xemc3.core.load.files)
        assert_ds_are_equal(dl, dn, True, 1e-4)


@settings(**setting)  # type: ignore
@given(gen_ds.hypo_shape(200), gen_ds.hypo_vars())
def test_write_load_some(shape, vars):
    ds = gen_ds.gen_rand(shape, vars)
    # if True:
    #    dir = "xemc3/test/test"
    with tempfile.TemporaryDirectory() as dir:
        xemc3.write.fortran(ds, dir)
        dl = xemc3.load(dir)
        assert_ds_are_equal(ds, dl, True, 1e-2, 1e-2)
    with tempfile.TemporaryDirectory() as dir:
        xemc3.write.fortran.all(dl, dir)
        dn = xemc3.load.all(dir)
        # print(xemc3.core.load.files)
        assert_ds_are_equal(dl, dn, True, 1e-4)


if __name__ == "__main__":
    test_write_load_simple()
    test_write_load_full()



import itertools
import sys

import mayavi.mlab as mlab  # type: ignore
import numpy as np
import xarray as xr
from tvtk.api import tvtk  # type: ignore

from .utils import rrange


def _pprint(*a, **k):
    print(*a, **k, end="")
    sys.stdout.flush()


class divertor:
    def __init__(
        self,
        ds,
        index,
        *,
        symmetry=False,
        segments=1,
        power_cutoff=None,
        verbose=False,
        vmax=None,
        vmin=None,
        only_lower=False,
        title=None,
        phi_slices=False,
        path1=None,
        path2=None,
    ):
        """
        Plot the first wall and the mapped plasma parameters.

        Parameters
        ----------
        ds : xr.Dataset
            The emc3 dataset of the mapped parameters
        index : str
            The index to plot
        symmetry : boolean
            Plot as well a copyied version of the data assuming
            stellerator symmetry.
        segments : int
            Plot several copies of the the data, spread around
            symmetrically.
        power_cutoff : float or None
            if not None, only plot target structures on which the
            total power flux on the structure is larger then the
            cutoff.
        verbose : boolean
            print informations during plotting
        vmax : None or float
            if not None set the upper limit of the colorbar
        vmin : None or float
            if not None set the lower limit of the colorbar
        only_lower : boolean
            only plot structures that are below :math:`z=0`
        title : str
            set a title for the plot
        phi_slices : boolean
            plot some additional debugging slices
        path1 : None or str
            Take a screenshot and save at location of str if not None
        path2 : None or str
            like path1 but screenshot is taken from different position
        """
        if vmax is None:
            vmax = np.nanmax(ds[index])
        if vmin is None:
            vmin = np.nanmin(ds[index])
        fig = mlab.figure(size=(1920, 1080))

        if verbose:
            _pprint("\nplotting")
        for plate in ds.emc3.iter_plates(symmetry=symmetry, segments=segments):
            if power_cutoff and power_cutoff > plate.tot_P:
                continue

            z = plate.emc3["z_corners"].data
            if only_lower:
                doit = False
                data = plate.emc3[index].data.copy()
                for i, j in rrange(data.shape):
                    if (z[i : i + 2, j : j + 2] > 0).all():
                        data[i, j] = np.nan
                    else:
                        doit = True
                if not doit:
                    continue
            else:
                data = plate.emc3[index].data
            r = plate.emc3["R_corners"].data
            phi = plate.emc3["phi_corners"].data
            if phi.shape != r.shape:
                phi = phi.reshape((z.shape[0], 1)) * np.ones((1, z.shape[1]))
            x = r * np.cos(phi)
            y = r * np.sin(phi)
            grid = [x, y, z]
            grid_centered = []
            ij = [(i % 2, i // 2) for i in range(4)]
            for s in grid:
                new = np.empty([(i - 1) * 2 for i in x.shape])
                slz = (slice(None, -1), slice(1, None))
                slz2 = (slice(None, None, 2), slice(1, None, 2))
                for i, j in ij:
                    new[slz2[i], slz2[j]] = s[slz[i], slz[j]]
                grid_centered.append(new)
            x, y, z = grid_centered
            datanew = np.empty_like(x)
            for i, j in ij:
                datanew[slz2[i], slz2[j]] = data
            msh = mlab.mesh(
                x, y, z, scalars=datanew, vmin=vmin, vmax=vmax, opacity=0.99
            )
            msh.module_manager.scalar_lut_manager.lut.nan_color = 0, 0, 0, 0
            if verbose:
                _pprint(".")
        if phi_slices:
            for phi in np.arange(np.pi * -0.2, np.pi * 0.2, np.pi / 180):
                R = np.array([[3.0, 6], [3, 6]])
                z = np.array([[1, 1], [-1, -1]])
                x = R * np.cos(phi)
                y = R * np.sin(phi)
                d = np.ones((2, 2))
                msh = mlab.mesh(x, y, z, scalars=d, opacity=0.99)
        if verbose:
            print(" done")
        if title:
            mlab.title(title, size=0.1, height=0.98)
        label = xr.plot.utils.label_from_attrs(ds[index])
        mlab.colorbar(title=label)
        # print("debug:",mlab.gcf() == fig)
        if path1:
            array = np.array
            mlab.view(
                -50,
                78.87131880429496,
                3.8,
                np.array([4.6, 0.17730785, -1.02087631]),
            )
            mlab.savefig(path1)  # , figure=self.fig)
            # print(mlab.screenshot())
        if path2:
            mlab.view(
                -28.62775839705135,
                64.3530947622391,
                2.6,
                array([4.28868816, -0.44711861, -1.2959868]),
            )
            # mlab.draw(fig)
            mlab.savefig(path2, figure=fig)  # , figure=self.fig)

        self.fig = fig
        # self.scene = scene
        # return self

    def savefig(self, path):
        return self

    def show(self):
        mlab.show()
        return self


class volume:
    def __init__(self, ds, updownsym=False, periodicity=1):
        print("__init__ started")
        print(updownsym, periodicity)
        self.ds = ds
        self.sym = updownsym
        self.period = periodicity
        vol_dims = [i[0] for i in [ds.r.shape, ds.theta.shape, ds.phi.shape]]
        # vol_dims[2] *= periodicity
        # if updownsym:
        #    vol_dims[2] *= 2
        print(vol_dims)
        self.dims = [i + 1 for i in vol_dims]
        self.grids = []
        mz = None
        z = ds.emc3["z_corners"].data.ravel(order="F")
        for i in range(self.period):
            ds["x_bounds"] = ds["R_bounds"] * np.cos(
                ds["phi_bounds"] + (2 * np.pi * i / self.period)
            )
            ds["y_bounds"] = ds["R_bounds"] * np.sin(
                ds["phi_bounds"] + (2 * np.pi * i / self.period)
            )
            x = ds.emc3["x_corners"].data.ravel(order="F")
            y = ds.emc3["y_corners"].data.ravel(order="F")
            self.grids.append(tvtk.StructuredGrid(dimensions=self.dims))
            self.grids[-1].points = np.ascontiguousarray(np.vstack([x, y, z]).T)
            if self.sym:
                if mz is None:
                    mz = -1 * z
                ds["x_bounds"] = ds["R_bounds"] * np.cos(
                    -ds["phi_bounds"] + (2 * np.pi * i / self.period)
                )
                ds["y_bounds"] = ds["R_bounds"] * np.sin(
                    -ds["phi_bounds"] + (2 * np.pi * i / self.period)
                )
                x = ds.emc3["x_corners"].data.ravel(order="F")
                y = ds.emc3["y_corners"].data.ravel(order="F")
                self.grids.append(tvtk.StructuredGrid(dimensions=self.dims))
                self.grids[-1].points = np.ascontiguousarray(np.vstack([x, y, mz]).T)

        # phi = ds["phi_bounds"]
        # phidims = phi.dims
        # phi = phi.data
        # R = ds["R_bounds"]
        # Rdims = R.dims
        # R = R.data
        # z = ds["z_bounds"].data
        # if self.sym:
        #     phi = np.append(-phi[::-1, ::-1], phi, axis=phidims.index("phi"))
        #     Rslc = [slice(None)] * 6
        #     Rslc[Rdims.index("phi")] = slice(None, None, -1)
        #     Rslc[Rdims.index("delta_phi")] = slice(None, None, -1)
        #     R = np.append(R[Rslc], R, axis=Rdims.index("phi"))
        #     z = np.append(-z[Rslc], z, axis=Rdims.index("phi"))
        # if self.period:
        #     phi0 = phi
        #     R0 = R
        #     z0 = z
        #     for i in range(self.period - 1):
        #         phi = np.append(
        #             phi,
        #             phi0 + (2 * np.pi) * (i + 1) / self.period,
        #             axis=phidims.index("phi"),
        #         )
        #         R = np.append(R, R0, axis=Rdims.index("phi"))
        #         z = np.append(z, z0, axis=Rdims.index("phi"))
        # dplot = xr.Dataset(
        #     coords=dict(
        #         R_bounds=(Rdims, R),
        #         z_bounds=(Rdims, z),
        #         phi_bounds=(phidims, phi),
        #     )
        # )
        print("__init__ finished")

    def plot(self, key, **kwargs):
        """
        Plot a some quantities in 3D

        Parameters
        ----------
        key : str
            The key to plot
        """
        print("plot started")
        # cell_data doesn't quite work ...
        # self.grid.cell_data.scalars = np.ascontiguousarray(self.ds[key].data).ravel(
        #    order="F"
        # )
        # self.grid.cell_data.scalars.name = "scalars"
        pnt_data = np.zeros(self.dims)
        fac = np.ones(self.dims)
        # rawdims = self.ds[key].dims
        rawdata = self.ds[key].data
        # if self.sym:
        #    rawdata = np.append(rawdata, rawdata, axis=rawdims.index("phi"))
        # rawdata0 = rawdata
        # for i in range(self.period - 1):
        #    rawdata = np.append(rawdata, rawdata0, axis=rawdims.index("phi"))
        for ijk in itertools.product(*[[slice(1, None), slice(None, -1)]] * 3):
            pnt_data[ijk] += rawdata
            fac[ijk] += 1
        pnt_data /= fac
        default = dict(opacity=0.9)
        default.update(kwargs)
        ml, mu = np.nanmin(pnt_data), np.nanmax(pnt_data)
        for i in range(self.period * (2 if self.sym else 2)):
            self.grids[i].point_data.scalars = pnt_data.ravel(order="F")
            self.grids[i].point_data.scalars.name = "scalars"
            d = mlab.pipeline.add_dataset(self.grids[i])
            # gx = mlab.pipeline.grid_plane(d)
            # gy = mlab.pipeline.grid_plane(d)
            # gy.grid_plane.axis = "y"
            # gz = mlab.pipeline.grid_plane(d)
            # gz.grid_plane.axis = "z"
            iso = mlab.pipeline.iso_surface(d, **default)
            # ml, mu = self.ds[key].min(), self.ds[key].max()
            iso.contour.maximum_contour = (mu - ml) / 2 + ml
        print("plot finished")
        return self

    def show(self):
        mlab.show()
        return self



import xarray as xr
import numpy as np
import re
import typing

from .utils import rrange


if 0:
    import sparse  # type: ignore
else:
    sparse = None

# #################################################
# Abstractions between sparse and full data:
# #################################################


def newarray(shape, dtype=None, fill_value=np.nan):
    if dtype == bool and fill_value is np.nan:
        fill_value = False
    if sparse:
        if dtype:
            return sparse.DOK(shape=shape, dtype=dtype, fill_value=fill_value)
        return sparse.DOK(shape=shape, fill_value=fill_value)
    if dtype:
        ret = np.empty(shape, dtype=dtype)
    else:
        ret = np.empty(shape)
    ret[:] = fill_value
    return ret


def tocoo(data):
    if sparse:
        return data.to_coo()
    return data


def nnz(data):
    if sparse:
        return data.nnz
    return np.sum(np.isfinite(data))


def keys(data):
    if hasattr(data, "coords"):
        for ijk in data.coords.T:
            yield ijk
        return
    try:
        for ijk in data.keys():
            yield ijk
        return
    except AttributeError:
        for ijk in rrange(data.shape):
            if np.isfinite(data[ijk]):
                yield ijk


def sparseCOO(locs, data, shape, fill_value):
    if sparse:
        return sparse.COO(locs, data, shape, fill_value)
    ret = newarray(shape, data.dtype, fill_value)
    retflat = ret.reshape(-1)
    for ijk, d in zip(locs, data):
        retflat[ijk] = d
    return ret


# #################################################
# The "main" code:
# #################################################


def read_depo_raw(ds: xr.Dataset, fn: str) -> typing.List[xr.DataArray]:
    bad = re.compile(r"(\d)([+-]\d)")
    # print(ds)

    shape = tuple([x + 1 for x in [len(ds.r), len(ds.theta), len(ds.phi)]])
    dims = tuple([x + "_plus1" for x in ["r", "theta", "phi"]])
    haszone = "zone" in ds.dims
    if haszone:
        shape = len(ds.zone), *shape
        dims = "zone", *dims

    with open(fn) as f:
        line = f.readline().split()
        assert len(line) == 2
        blockasize = int(line[0])
        # Lines look like this:
        # Index
        # 3
        # type of the surface
        # index of zone
        # index of surface in radial
        #                  in poloidal
        #                  in toroidal direction
        # flux
        out = {
            "surftype": newarray(shape=shape, dtype=bool, fill_value=True),
            "flux": newarray(shape=shape, fill_value=np.nan),
            "other": [newarray(shape=shape, fill_value=np.nan) for _ in range(4)],
        }
        varnames = "surftype", "flux", "other"

        hasother = hasother2 = False
        last = None
        for i in range(blockasize):
            line = bad.sub(r"\1E\2", f.readline()).split()
            ints = [int(x) for x in line[:7]]
            floats = [float(x) for x in line[7:]]
            assert (
                ints[0] == i + 1
            ), f"Expected first index to be contigous, thus expected {i+1} but got {ints[0]}"
            assert ints[1] == 3, f"Expected 3 but got {ints[1]}"
            # print(ints[3:])
            if haszone:
                slc = tuple(ints[3:])
            else:
                assert ints[3] == 0, f"Expected zoneid=0 but got {ints[3]}"
                slc = tuple(ints[4:])

            for name, val in zip(varnames, (ints[2] == 1, floats[0])):
                out[name][slc] = val
            if last:
                assert not out["surftype"][last]
            if ints[2] != 1:
                last = slc
            # print("a", out["surftype"][slc], ints[2] == 1)

            if len(floats) > 1:
                hasother = True
                for j in range(4):
                    out["other"][j][slc] = floats[j + 1]

        _ = f.readline()

        out2 = {
            "surftype": newarray(shape=shape, dtype=bool, fill_value=False),
            "flux": newarray(shape=shape, fill_value=np.nan),
            "other": [newarray(shape=shape, fill_value=np.nan) for _ in range(4)],
        }

        while True:
            line = bad.sub(r"\1E\2", f.readline()).split()
            if line == []:
                break
            ints = [int(x) for x in line[:7]]
            floats = [float(x) for x in line[7:]]
            assert ints[1] == 3, f"Expected 3 but got {ints[1]}"
            if haszone:
                slc = tuple(ints[3:])
            else:
                assert ints[3] == 0
                slc = tuple(ints[4:])
            for name, val in zip(varnames, (ints[2] == 1, floats[0])):
                out2[name][slc] = val
            # print("b", out["surftype"][slc], ints[2] == 1)

            if len(floats) > 1:
                hasother2 = True
                for j in range(4):
                    out2["other"][j][slc] = floats[j + 1]

    if sparse:
        assert any([d.nnz for d in out["other"]]) == hasother
        assert any([d.nnz for d in out2["other"]]) == hasother2
    # if not hasother:
    # out["other"] = []
    assert not hasother2
    out2["other"] = []

    ret = [
        xr.DataArray(data=tocoo(d), dims=dims)
        for d in [out["surftype"], out["flux"], *out["other"]]
    ], [
        xr.DataArray(data=tocoo(d), dims=dims)
        for d in [out2["surftype"], out2["flux"], *out2["other"]]
    ]
    assert len(ret[0]) == 6, f"{len(ret[0])}"
    assert len(ret[1]) == 2
    return ret[0] + ret[1]


def write_depo_raw_part(datas, f, i):
    surf = datas[0].data
    dats = [d.data for d in datas[1:]]
    zone = [] if "zone" in datas[0].dims else [0]
    off = 1 if i == 0 else 0
    # print(keys(dats[0]))
    for ijk in keys(dats[0]):
        i += off
        ijk = tuple(ijk)
        s = 1 if surf[ijk] else -1

        zijk = *zone, *ijk
        f.write(
            f"{i:6d} 3 {s:2d} {zijk[0]:2d} "
            + (
                " ".join(
                    [f"{d:4d}" for d in zijk[1:]] + [f"{d[ijk]:11.4E}" for d in dats]
                )
            )
            + "\n"
        )
    return i


def write_depo_raw(datas, fn):
    assert len(datas) == 8, f"Expected 8 data entries, but got { len(datas) }"
    datas = datas[:6], datas[6:]
    i = 0
    with open(fn, "w") as f:
        f.write(
            f"      {nnz(datas[0][1].data):6d} "
            + (
                f"{nnz(datas[1][1].data):11d}\n"
                if "PARTICLE_DEPO" in fn
                else "TARGET\n"
            )
        )
        # for k in datas[0]:
        #     print()
        #     print(k)
        i = write_depo_raw_part(datas[0], f, i)
        f.write(" MAPPING\n")
        write_depo_raw_part(datas[1], f, i + 1)


# #################################################
# Code for writing sparse files. Currently not used
# #################################################


def da_to_netcdf(da, fn):
    if hasattr(da.data, "nnz"):
        ds = xr.Dataset()
        ds["var"] = da
        return ds_to_netcdf(ds, fn)
    da.to_netcdf()


def ds_to_netcdf(ds, fn):
    dsorg = ds
    ds = dsorg.copy()
    for v in ds:
        if hasattr(ds[v].data, "nnz") and (
            hasattr(ds[v].data, "to_coo") or hasattr(ds[v].data, "linear_loc")
        ):
            coord = f"_{v}_xarray_index_"
            assert coord not in ds
            data = ds[v].data
            if hasattr(data, "to_coo"):
                data = data.to_coo()
            ds[coord] = coord, data.linear_loc()
            dims = ds[v].dims
            ds[coord].attrs["compress"] = " ".join(dims)
            at = ds[v].attrs
            ds[v] = coord, data.data
            ds[v].attrs = at
            ds[v].attrs["_fill_value"] = str(data.fill_value)
            for d in dims:
                if d not in ds:
                    ds[f"_len_{d}"] = len(dsorg[d])

    # print(ds)
    ds.to_netcdf(fn)


def xr_open_dataset(fn):
    ds = xr.open_dataset(fn)

    def fromflat(shape, i):
        index = []
        for fac in shape[::-1]:
            index.append(i % fac)
            i //= fac
        return tuple(index[::-1])

    for c in ds.coords:
        if "compress" in ds[c].attrs:
            vs = c.split("_")
            if len(vs) < 5:
                continue
            if vs[-1] != "" or vs[-2] != "index" or vs[-3] != "xarray":
                continue
            v = "_".join(vs[1:-3])
            # at = ds[v].attrs
            dat = ds[v].data
            fill = ds[v].attrs.pop("_fill_value", None)
            if fill:
                knownfails = {"nan": np.nan, "False": False, "True": True}
                if fill in knownfails:
                    fill = knownfails[fill]
                else:
                    fill = np.fromstring(fill, dtype=dat.dtype)
            dims = ds[c].attrs["compress"].split()
            shape = []
            for d in dims:
                try:
                    shape.append(len(ds[d]))
                except KeyError:
                    shape.append(int(ds[f"_len_{d}"].data))
                    ds = ds.drop_vars(f"_len_{d}")

            locs = fromflat(shape, ds[c].data)
            data = sparseCOO(locs, ds[v].data, shape, fill_value=fill)
            ds[v] = dims, data, ds[v].attrs, ds[v].encoding
    # print(ds)
    return ds


# #################################################
# Code for manual testing ...
# #################################################


def get_ds():
    cache = "this.nc"
    try:
        return xr.open_dataset(cache)
    except FileNotFoundError:
        pass
    import xemc3

    ds = xemc3.load(".")
    ds.to_netcdf(cache)
    return ds


def get_p(ds):
    cache = "test_p.nc"
    try:
        ds = xr_open_dataset(cache)
    except FileNotFoundError:
        pass
    else:
        x = [[ds[f"var{j}_{i}"] for i in range(6 - j * 4)] for j in range(2)]
        return x

    x = read_depo_raw(ds, "PARTICLE_DEPO")
    ds = xr.Dataset()
    for i, t in enumerate(x):
        for j, v in enumerate(t):
            ds[f"var{i}_{j}"] = v
    ds_to_netcdf(ds, cache)
    return x


def get_e(ds):
    cache = "test_e.nc"
    try:
        ds = xr_open_dataset(cache)
    except FileNotFoundError:
        pass
    else:
        x = [[ds[f"var{j}_{i}"] for i in range(6 - j * 4)] for j in range(2)]
        return x

    x = read_depo_raw(ds, "ENERGY_DEPO")
    ds = xr.Dataset()
    for i, t in enumerate(x):
        for j, v in enumerate(t):
            ds[f"var{i}_{j}"] = v
    ds_to_netcdf(ds, cache)
    return x


if __name__ == "__main__":
    ds = get_ds()

    x = get_p(ds)
    t = x[0][1:]
    # print(len(t))
    # print(xr.combine_nested(t, "time"))

    write_depo_raw(x, "PARTICLE_DEPO_NEW")
    y = get_e(ds)

    # read_depo_raw(ds, "ENERGY_DEPO")
    write_depo_raw(y, "ENERGY_DEPO_NEW")



"""Routines for reading EMC3 files."""

__all__ = ["plates", "mappings", "all", "mapped_raw", "file", "var", "plates_geom"]

import sys as _sys
import types as _types

# Actual implementation is in load.py
from ..core.load import (
    get_plates as plates,
    load_all as all,
    load_any as any,
    read_fort_file_pub as file,
    read_mapped as mapped_raw,
    read_plate_nice as plates_geom,
    read_var as var,
)


class _any(_types.ModuleType):
    def __init__(self):
        super().__init__(__name__)
        self.__dict__.update(_sys.modules[__name__].__dict__)

    def __call__(self, path, ignore_missing=None):
        return any(path, ignore_missing)


_any.__call__.__doc__ = any.__doc__

_sys.modules[__name__] = _any()



import os

import matplotlib.pyplot as plt

if "DOCUTILSCONFIG" in os.environ:
    get_ipython().run_line_magic("matplotlib", "inline")
    figsize = 10.0
else:
    get_ipython().run_line_magic("matplotlib", "notebook")
    figsize = 6.0
plt.rcParams["figure.figsize"] = [figsize, figsize]



import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import xarray as xr

from . import utils
from .load import plate_prefix


def plot_rz(
    ds,
    key,
    phi,
    ax=None,
    Rmin=None,
    Rmax=None,
    zmin=None,
    zmax=None,
    aspect=True,
    figsize=None,
    target=False,
    **kwargs,
):

    phis = ds["phi_bounds"]
    if phi < np.min(phis.data) or phi > np.max(phis.data):
        raise RuntimeError(
            f"{phi} outside of bounds in dataset {np.min(phis)}:{np.max(phis)}"
        )
    for phi_i, phib in enumerate(phis):
        if phib[0] <= phi <= phib[1]:
            break
    else:
        raise RuntimeError(f"no suitable phi slice found for {phi} in {phis}")

    p = ((phi - phib[0]) / (phib[1] - phib[0])).data
    ds = ds.isel(phi=phi_i)
    das = [ds[k] for k in ["R_bounds", "z_bounds"]]
    if key:
        das.append(ds[key])
    pp = xr.DataArray(data=[(1 - p), p], dims="delta_phi")
    das = [
        (da * pp).sum(dim="delta_phi") if "delta_phi" in da.dims else da for da in das
    ]
    if key:
        data = das[2].data
        if "time" in ds[key].dims:
            raise ValueError(
                "Unexpected dimension `time` - animation is not yet supported!"
            )
        if len(ds[key].dims) != 2:
            raise ValueError(
                f"Expected 2 dimensions for R-z plot, but found {len(ds[key].dims)}: {ds[key].dims}!"
            )
    else:
        data = np.zeros(das[0].shape[:2]) * np.nan
        if "edgecolors" not in kwargs:
            kwargs["edgecolors"] = "k"
    r = utils.from_interval(das[0])
    z = utils.from_interval(das[1])
    ax = _get_ax(figsize, ax)
    p = ax.pcolormesh(r, z, data, **kwargs)
    # plt.xlabel(xr.plot.utils.label_from_attrs(r))
    if aspect:
        ax.set_aspect(1)
    ax.set_xlabel("R [m]")
    ax.set_ylabel("z [m]")
    plt.colorbar(p, ax=ax)
    if Rmin is not None or Rmax is not None:
        ax.set_xlim(Rmin, Rmax)
    if zmin is not None or zmax is not None:
        ax.set_ylim(zmin, zmax)
    p.colorbar.set_label(label=xr.plot.utils.label_from_attrs(das[-1]))
    if target:
        plot_target(ds, phi, ax=ax, fmt="r-" if key is None else "k-", aspect=aspect)
    return p


def plot_target(ds, phi, fmt=None, ax=None, figsize=None, aspect=True):
    ax = _get_ax(figsize, ax)
    if aspect:
        ax.set_aspect(1)
    das = xr.Dataset()
    for k in "R", "phi", "z":
        k1 = f"{plate_prefix}{k}"
        das[k] = ds[k1]
    for k in ds:
        if k.endswith("_dims") and k.startswith(f"_{plate_prefix}"):
            das[k] = ds[k]
    for p in das.emc3.iter_plates(symmetry=True, segments=5):
        assert len(p.plate_phi.shape) == 1
        if not (p.plate_phi.min() <= phi <= p.plate_phi.max()):
            continue
        p["plate_phi_plus1"] = p.plate_phi
        p = p.interp(
            plate_phi_plus1=phi, assume_sorted=False, kwargs=dict(bounds_error=True)
        )
        # except ValueError as e:
        #    assert "value in x_new is" in e.args[0]
        #    continue

        assert np.isclose(p.plate_phi, phi), f"{phi} expected but got {p.plate_phi}"
        ax.plot(p.plate_R, p.plate_z, fmt or "k-")

    ax.set_xlabel("R [m]")
    ax.set_ylabel("z [m]")


def _get_ax(figsize, ax):
    if figsize is not None:
        assert (
            ax is None
        ), "Passing in an axes object and specifing the figure size cannot be combined"
        plt.figure(figsize=figsize)
    if ax is None:
        ax = plt.axes(label=np.random.bytes(20))
    return ax



import os


def get_data(force=False):
    basedir = "./example-data/"
    if not os.path.isdir(basedir) and not force:
        pytest.skip("create {basedir} to enable testing on real data")
        return
    if not os.path.isdir(basedir + ".git"):
        os.system(
            f"git clone https://oauth2:uQ7_vw_m5fo4UFQrenUs@gitlab.mpcdf.mpg.de/dave/xemc3-data/ {basedir} --depth 1"
        )
    else:
        os.system(f"cd {basedir}; git fetch origin main --depth 1")
        os.system(f"cd {basedir}; git checkout origin main")
    return basedir + "emc3_example"


if __name__ == "__main__":
    import sys

    get_data(force=True)
    sys.exit(0)


import numpy as np
import pytest
import xarray as xr
from xarray.testing import assert_identical

import xemc3

# simple regression test
def test_load_all():
    bd = get_data()
    result = xemc3.load.all(bd)
    expected = xr.open_dataset(bd + ".nc")

    # Remove changing attributes
    changing = "software_version", "date_created", "id"
    for a in changing:
        assert a in result.attrs
        result.attrs.pop(a, None)
        expected.attrs.pop(a, None)

    # Remove new attributes, so we don't have to regenerate the data that often
    for k in list(result) + list(result.coords):
        if k not in expected:
            del result[k]
            continue
        for a in list(result[k].attrs):
            if a not in expected[k].attrs:
                del result[k].attrs[a]
    for a in list(result.attrs):
        if a not in expected.attrs:
            del result.attrs[a]

    assert_identical(result, expected)


def test_iter_plates():
    bd = get_data()
    ds = xemc3.load.plates(bd)
    c = 0
    for x in ds.emc3.iter_plates():
        c += 1
        assert np.all(np.isfinite(x.emc3["f_E"].data))
    assert c == 22
    c = 0
    for _ in ds.emc3.iter_plates(symmetry=True, segments=5):
        c += 1
    assert c == 22 * 5 * 2


def test_get_element():
    bd = get_data()
    ds = xemc3.load.plates(bd)
    fe = ds.emc3["f_E"]
    assert isinstance(fe, list)
    for fel in fe:
        assert isinstance(fel, xr.DataArray)
        assert fel.attrs == ds["f_E"].attrs
